import base64
import enum
import json
import logging
import re
import urllib.parse
from dataclasses import dataclass
from decimal import Decimal
from os import fdopen
from typing import Any, Dict, List, Optional, Union

import base58
import bdkpython as bdk

from bitcoin_qr_tools.converter_xpub import ConverterXpub
from bitcoin_qr_tools.i18n import translate
from bitcoin_qr_tools.signer_info import SignerInfo
from bitcoin_qr_tools.utils import (
    DecodingException,
    InconsistentDescriptors,
    InvalidBitcoinURI,
    WrongNetwork,
    serialized_to_hex,
)

from .multipath_descriptor import MultipathDescriptor

BITCOIN_BIP21_URI_SCHEME = "bitcoin"
logger = logging.getLogger(__name__)


class DataType(enum.Enum):
    Bip21 = enum.auto()  # https://bips.dev/21/
    Descriptor = enum.auto()
    MultiPathDescriptor = enum.auto()
    Xpub = enum.auto()
    Fingerprint = enum.auto()
    SignerInfo = enum.auto()  # FingerPrint, Derivation path, Xpub
    PSBT = enum.auto()
    Txid = enum.auto()
    Tx = enum.auto()
    SignerInfos = (
        enum.auto()
    )  # a list of SignerInfo with matching fingerprints (fingerprint, means here the root fingerprint)
    LabelsBip329 = enum.auto()
    UnrelatedSignerInfos = (
        enum.auto()
    )  # a list of SignerInfo, that do not (necessarily share the root fingerprint)
    MultisigWalletExport = enum.auto()

    @classmethod
    def from_value(cls, value: int) -> "DataType":
        min_value = min([member.value for member in cls])
        return list(cls)[value - min_value]

    @classmethod
    def from_name(cls, value: str) -> "DataType":
        names = [member.name for member in cls]
        return list(cls)[names.index(value)]


class ConverterTools:
    @classmethod
    def base43_decode(cls, v: str):
        "here is the simplified electrum base43 code"
        __b43chars = b"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ$*+-./:"
        assert len(__b43chars) == 43
        __b43chars_inv = {v: k for k, v in enumerate(__b43chars)}

        base = 43
        num = 0

        # Remove leading zeros and adjust length
        v = v.lstrip("0")

        # Convert each character to a number using the inverse character map
        for char in v:
            num = num * base + __b43chars_inv[ord(char)]

        # Convert the number to bytes
        return num.to_bytes((num.bit_length() + 7) // 8, "big")

    @classmethod
    def _decoding_strategies(cls):
        return [
            lambda x: base64.b64decode(x),  # base64 decoding
            lambda x: bytes.fromhex(x),  # hex decoding
            lambda x: cls.base43_decode(x),  # base43 decoding
            lambda x: base58.b58decode(x),
        ]


class ConverterTxid:
    @classmethod
    def is_valid_bitcoin_hash(cls, hash: str) -> bool:
        import re

        if re.match("^[a-f0-9]{64}$", hash):
            return True
        else:
            return False


class ConverterFingerprint:
    @classmethod
    def is_valid_wallet_fingerprint(cls, fingerprint: str) -> bool:
        import re

        if re.match("^[a-fA-F0-9]{8}$", fingerprint):
            return True
        else:
            return False


class ConverterAddress:
    @classmethod
    def is_bitcoin_address(cls, s, network: bdk.Network):
        try:
            bdkaddress = bdk.Address(s, network)
            return bool(bdkaddress) and bdkaddress.is_valid_for_network(network=network)
        except:
            return False


@dataclass
class ConverterMultisigWalletExport:
    name: str
    threshold: int
    address_type_short_name: str
    signer_infos: List[SignerInfo]

    def to_str(self) -> str:
        return self.to_custom_str(hardware_signer_name="")

    def to_custom_str(self, hardware_signer_name="Passport") -> str:
        return f"""# {hardware_signer_name} Multisig setup file (created by Bitcoin Safe)
#
Name: {self.name}
Policy: {self.threshold} of {len(self.signer_infos)}
Format: {self.address_type_short_name.upper()}

""" + "\n".join(
            [
                f"Derivation: {spk_provider.key_origin}\n{spk_provider.fingerprint}: {spk_provider.xpub}"
                for spk_provider in self.signer_infos
            ]
        )

    @classmethod
    def parse_from_legacy_coldcard(
        cls, s: str, network: bdk.Network
    ) -> "Optional[ConverterMultisigWalletExport]":
        """

        Can parse a string like
            # Exported by Blockstream Jade
            Name: hwi3374c2e55c4b
            Policy: 2 of 3
            Format: P2WSH
            Derivation: m/48'/1'/0'/2'
            14c949b4: tpubDDvtDSGt5JmgxgpRp3nyZj3ULZvFWuU9AaS6x3UwkNE6vaNgzd6oyKYEQUzSevUQs2ste5QznpbN8Nt5bVbZvrJFpCqw9UPXCtnCutEvEwW
            Derivation: m/48'/1'/0'/2'
            d8cf7475: tpubDEDUiUcwmoC92QJ2kGPQwtikGqLrjdyUfuRMhm5ab4nYmgRkkKPF9mp2FcunzMu9y5Ea2urGUJh4t1o7Wb6KjKddzJKcE8BoAyTWK6ughFK
            Derivation: m/48'/1'/0'/2'
            d5b43540: tpubDFnCcKU3iUF4sPeQC68r2ewDaBB7TvLmQBTs12hnNS8nu6CPjZPmzapp7Woz6bkFuLfSjSpg6gacheKBaWBhDnEbEpKtCnVFdQnfhYGkPQF"


        Args:
            s (str): _description_
            network (bdk.Network): _description_

        Returns:
            Optional[ConverterWalletExport]: _description_
        """
        lines = s.split("\n")

        def extract_value(line: str, key: str) -> Optional[str]:
            if line.startswith(key):
                return line[len(key) :].strip()
            return None

        def extract_unique(key: str) -> Optional[str]:
            for line in lines:
                res = extract_value(line, key=key)
                if res:
                    return res
            return None

        name = extract_unique("Name:")
        if not name:
            return None

        policy = extract_unique("Policy:")
        if not policy:
            return None
        if not " of " in policy:
            return None
        policy_parts = policy.split(" of ")
        if len(policy_parts) != 2:
            return None
        _threshold, _num_signers = policy_parts
        try:
            threshold = int(_threshold.strip())
            num_signers = int(_num_signers.strip())
        except:
            return None
        if not (0 < threshold <= num_signers):
            return None

        format = extract_unique("Format:")
        if not format:
            return None
        address_type_short_name = format.strip().lower()

        current_derivation = None
        # get the signer_infos
        signer_infos: List[SignerInfo] = []
        for line in lines:
            if line.startswith("#"):
                continue

            # if it is a derivation line
            if this_derivation := extract_value(line, "Derivation:"):
                current_derivation = this_derivation

            # if it is a fingerprint
            if not ":" in line:
                continue
            parts = line.split(":")
            if len(parts) != 2:
                continue
            first, last = parts
            if not ConverterFingerprint.is_valid_wallet_fingerprint(first):
                continue

            fingerprint = first.strip()

            last = last.strip()
            if not ConverterXpub.is_xpub(last):
                continue
            if not ConverterXpub.xpub_matches_network(last, network=network):
                raise WrongNetwork(f"xpub doesnt match network {network.name}")
            xpub = last

            #  ensure that current_derivation is set
            if not current_derivation:
                continue

            signer_infos.append(SignerInfo(xpub=xpub, fingerprint=fingerprint, key_origin=current_derivation))

        if not signer_infos or num_signers != len(signer_infos):
            return None

        return ConverterMultisigWalletExport(
            name=name,
            threshold=threshold,
            address_type_short_name=address_type_short_name,
            signer_infos=signer_infos,
        )


class ConverterBip329:
    @classmethod
    def is_ndjson_with_keys(cls, s: str, keys: List[str]):
        """
        Checks if the input string s is newline-delimited JSON and each JSON object contains the specified keys.

        Args:
        s (str): The input string to check.
        keys (list of str): The keys that must be present in each JSON object.

        Returns:
        bool: True if the string is newline-delimited JSON and each object contains the specified keys, False otherwise.
        """
        lines = s.splitlines()

        # Check if there's at least one line
        if not lines:
            return False

        for line in lines:
            try:
                obj = json.loads(line)
                # Check if all specified keys are present in the JSON object
                if not all(key in obj for key in keys):
                    logger.debug(f"ndjson check: Not all required keys {keys} are present in {obj}")
                    return False
            except json.JSONDecodeError:
                return False

        return True

    @classmethod
    def is_bip329(cls, s: str) -> bool:
        try:
            return cls.is_ndjson_with_keys(s, keys=["type", "ref", "label"])
        except:
            return False


class ConverterBip21:
    def __init__(self, data: Dict) -> None:
        self.data = data

    def to_json(self):
        return json.dumps(self.data)

    @classmethod
    def decode_bip21_uri(cls, uri: str, network: bdk.Network) -> dict:
        """Raises InvalidBitcoinURI on malformed URI."""
        TOTAL_COIN_SUPPLY_LIMIT_IN_BTC = 21000000
        COIN = 100000000

        if not isinstance(uri, str):
            raise InvalidBitcoinURI(f"expected string, not {repr(uri)}")

        if ":" not in uri:
            if ConverterAddress.is_bitcoin_address(uri, network=network):
                return {"address": uri}
            else:
                raise InvalidBitcoinURI("Not a bitcoin address")

        u = urllib.parse.urlparse(uri)
        if u.scheme.lower() != BITCOIN_BIP21_URI_SCHEME:
            raise InvalidBitcoinURI("Not a bitcoin URI")
        address = u.path

        # python for android fails to parse query
        if address.find("?") > 0:
            address, query = u.path.split("?")
            pq = urllib.parse.parse_qs(query)
        else:
            pq = urllib.parse.parse_qs(u.query)

        for k, v in pq.items():
            if len(v) != 1:
                raise InvalidBitcoinURI(f"Duplicate Key: {repr(k)}")

        out: Dict[str, Any] = {k: v[0] for k, v in pq.items()}
        if address:
            if not ConverterAddress.is_bitcoin_address(address, network=network):
                raise InvalidBitcoinURI(f"Invalid bitcoin address: {address}")
            out["address"] = address
        if "amount" in out:
            am = out["amount"]
            try:
                m = re.match(r"([0-9.]+)X([0-9])", am)
                if m:
                    amount = Decimal(m.group(1)) * pow(Decimal(10), int(m.group(2)) - 8)
                else:
                    amount = Decimal(am) * COIN
                if amount > TOTAL_COIN_SUPPLY_LIMIT_IN_BTC * COIN:
                    raise InvalidBitcoinURI(f"amount is out-of-bounds: {amount!r} BTC")
                out["amount"] = int(amount)
            except Exception as e:
                raise InvalidBitcoinURI(f"failed to parse 'amount' field: {repr(e)}") from e
        if "message" in out:
            out["message"] = out["message"]
            out["memo"] = out["message"]
        if "time" in out:
            try:
                out["time"] = int(out["time"])
            except Exception as e:
                raise InvalidBitcoinURI(f"failed to parse 'time' field: {repr(e)}") from e
        if "exp" in out:
            try:
                out["exp"] = int(out["exp"])
            except Exception as e:
                raise InvalidBitcoinURI(f"failed to parse 'exp' field: {repr(e)}") from e
        if "sig" in out:
            try:
                out["sig"] = serialized_to_hex(out["sig"])
            except Exception as e:
                raise InvalidBitcoinURI(f"failed to parse 'sig' field: {repr(e)}") from e

        return out

    @classmethod
    def _try_decode_bip21(cls, s, network: bdk.Network):
        try:
            return cls.decode_bip21_uri(s, network=network)
        except Exception:
            pass


class ConverterTx:
    def __init__(self, tx: bdk.Transaction) -> None:
        self.tx = tx

    def to_str(self):
        return str(serialized_to_hex(self.tx.serialize()))

    @classmethod
    def _try_transaction_binary(cls, raw: bytes) -> Optional[bdk.Transaction]:
        # Try each decoding strategy in the loop
        try:
            return bdk.Transaction(list(raw))
        except Exception:
            return None

    @classmethod
    def _try_decode_serialized_transaction(cls, s: str) -> Optional[bdk.Transaction]:
        # Try each decoding strategy in the loop
        for decode in ConverterTools._decoding_strategies():
            try:
                decoded = decode(s)
                return bdk.Transaction(list(decoded))
            except Exception:
                continue  # If one strategy fails, try the next

        return None


class ConverterPSBT:
    def __init__(self, psbt: bdk.PartiallySignedTransaction) -> None:
        self.psbt = psbt

    def to_str(self):
        return str(self.psbt.serialize())

    @classmethod
    def _try_decode_psbt_binary(cls, raw: bytes) -> Optional[bdk.PartiallySignedTransaction]:
        psbt_magic_bytes = b"psbt\xff"

        # Try each decoding strategy in the loop
        try:
            if raw[: len(psbt_magic_bytes)] == psbt_magic_bytes:
                return bdk.PartiallySignedTransaction(base64.b64encode(raw).decode())
        except Exception:
            return None
        return None

    @classmethod
    def _try_decode_psbt(cls, s) -> Optional[bdk.PartiallySignedTransaction]:
        psbt_magic_bytes = b"psbt\xff"

        # Try each decoding strategy in the loop
        for decode in ConverterTools._decoding_strategies():
            try:
                decoded = decode(s)
                if decoded[:5] == psbt_magic_bytes:
                    return bdk.PartiallySignedTransaction(base64.b64encode(decoded).decode())
            except Exception:
                continue  # If one strategy fails, try the next

        return None


class ConverterSignerInfo:
    def __init__(self, info: SignerInfo) -> None:
        self.data = info

    def to_json(self):
        return self.data.to_json()

    @classmethod
    def _try_extract_signer_info(cls, s: str) -> Optional[SignerInfo]:
        signer_info = None
        try:
            signer_info = SignerInfo.from_str(s)
        except:
            pass

        if signer_info:
            signer_info.xpub = ConverterXpub.normalized_to_bip32(signer_info.xpub)
            return signer_info

        # try to load from a generic json (and cobo)
        try:
            d = json.loads(s)
            fingerprint = None
            key_origin = None
            xpub = None
            derivation_path = None
            if d.get("fingerprint"):
                fingerprint = d.get("fingerprint")
            if d.get("xfp"):
                fingerprint = d.get("xfp")
            if d.get("key_origin"):
                key_origin = d.get("key_origin")
            if d.get("deriv"):
                key_origin = d.get("deriv")
            if d.get("path"):
                # cobo convention
                key_origin = d.get("path")
            if d.get("xpub"):
                xpub = d.get("xpub")
            if d.get("derivation_path"):
                derivation_path = d.get("derivation_path")
            if fingerprint and key_origin and xpub:
                return SignerInfo(
                    fingerprint=fingerprint, key_origin=key_origin, xpub=xpub, derivation_path=derivation_path
                )
        except Exception:
            pass

        return None


class ConverterSignerInfos:
    def __init__(self, infos: List[SignerInfo], network: bdk.Network) -> None:
        self.data = infos
        self.network = network

    @classmethod
    def _try_multisig_wallet_export(
        cls, s: str, network: bdk.Network
    ) -> Optional[ConverterMultisigWalletExport]:
        """_summary_

        Args:
            s (str): _description_
            network (bdk.Network): _description_

        Raises:
            e: WrongNetwork, if the format is correct, but it is WrongNetwork

        Returns:
            Optional[ConverterMultisigWalletExport]: _description_
        """
        try:
            return ConverterMultisigWalletExport.parse_from_legacy_coldcard(s, network=network)
        except Exception as e:
            if isinstance(e, WrongNetwork):
                raise e
        return None

    def sparrow_format(self) -> str:
        first_signer_info = self.data[0]
        assert isinstance(first_signer_info, SignerInfo)

        d: Dict[str, Any] = {
            "chain": "BTC"
            if self.network
            in [
                bdk.Network.BITCOIN,
            ]
            else "XRT",
            "xfp": first_signer_info.fingerprint,  # root fingerprint
        }

        for i, signer_info in enumerate(self.data):
            assert isinstance(signer_info, SignerInfo)
            assert first_signer_info.fingerprint == signer_info.fingerprint, translate(
                "data",
                "The fingerprints differ.  Only same fingerprints are supported, ensuring all derived keys belong to the same signer!",
            )

            key = signer_info.name if signer_info.name and (signer_info.name not in d) else str(i)
            d[key] = {
                "xpub": signer_info.xpub,
                # "xfp" :  do not give here the fingerprint, because the the sparrow format assumes the fingerprint at this derived key, rather than the root_fingerprint
                "deriv": signer_info.key_origin,
                "first": signer_info.first_address,
                "name": signer_info.name,
            }
        return json.dumps(d)

    @classmethod
    def _try_extract_sparrow_signer_infos(cls, s, network: bdk.Network) -> Optional[List[SignerInfo]]:
        # if it is a json
        json_data = None
        try:
            json_data = json.loads(s)

            # check if it is in sparrow export format
            assert "chain" in json_data
            assert "xfp" in json_data
            # assert "xpub" in json_data  # this is not necessarily always known
        except:
            return None

        if network == bdk.Network.BITCOIN:
            if json_data["chain"] != "BTC":
                raise WrongNetwork(f"""Expected Network {network}, but got {json_data["chain"]}""")
        if network == bdk.Network.REGTEST:
            if json_data["chain"] not in ["XRT", "TBTC"]:  # XTR is coinkite
                raise WrongNetwork(f"""Expected Network {network}, but got {json_data["chain"]}""")
        if network == bdk.Network.TESTNET:
            if json_data["chain"] not in ["XTN", "TBTC"]:  # XTN is coinkite
                raise WrongNetwork(f"""Expected Network {network}, but got {json_data["chain"]}""")
        if network == bdk.Network.SIGNET:
            # unclear which chain value is used for signet in coldcard
            # https://coldcard.com/docs/upgrade/#mk4-version-511-feb-27-2023
            if json_data["chain"] not in ["XTN", "XRT", "TBTC"]:
                raise WrongNetwork(f"""Expected Network {network}, but got {json_data["chain"]}""")

        # the fingerprint in the top level is the relevant one
        fingerprint = json_data["xfp"]

        return [
            SignerInfo(
                xpub=v["xpub"],
                fingerprint=fingerprint,
                key_origin=v["deriv"],
                first_address=v["first"] if "first" in v else None,
                name=v["name"],
            )
            for v in json_data.values()
            if isinstance(v, dict)
        ]

    @classmethod
    def _try_extract_multisig_signer_infos_coldcard_and_passport_qr(
        cls, s, network: bdk.Network
    ) -> Optional[List[SignerInfo]]:
        # if it is a json
        json_data = None
        try:
            json_data = json.loads(s)

            # check if it is in coldcard export format
            assert "xfp" in json_data
            # for coldcard  assert "account" in json_data
            # for passport  assert "account" not in json_data

        except Exception:
            return None

        # the fingerprint in the top level is the relevant one
        fingerprint = json_data["xfp"]

        # get all the main keys (address type names) by looking at which end with _deriv
        address_type_names = [key.rstrip("_deriv") for key in json_data if key.endswith("_deriv")]

        signer_infos = [
            SignerInfo(
                xpub=ConverterXpub.normalized_to_bip32(json_data[address_type_name]),
                fingerprint=fingerprint,
                key_origin=json_data[address_type_name + "_deriv"],
                name=address_type_name,
            )
            for address_type_name in address_type_names
        ]
        for signer_info in signer_infos:
            ConverterXpub.ensure_xpub_matches_network(signer_info.xpub, network=network)
        return signer_infos


class ConverterDescriptor:
    def __init__(self, descriptor: Union[bdk.Descriptor, MultipathDescriptor]) -> None:
        self.descriptor = descriptor

    def to_str(self):
        return self.descriptor.as_string_private()

    @classmethod
    def _try_get_descriptor(cls, s, network: bdk.Network) -> Optional[bdk.Descriptor]:
        try:
            assert "<" not in s, "This contains characters of a multipath descriptor"
            assert ">" not in s, "This contains characters of a multipath descriptor"
            descriptor = bdk.Descriptor(s, network)
            if descriptor:
                logger.debug("detected descriptor")
                return descriptor
        except Exception:
            pass

        try:
            specter_dict = json.loads(s)
            if "descriptor" in specter_dict:
                return cls._try_get_descriptor(specter_dict["descriptor"], network=network)
        except Exception:
            pass
        return None

    @classmethod
    def _try_get_multipath_descriptor(cls, s, network: bdk.Network) -> Optional[MultipathDescriptor]:
        # if new lines are presnt, try checking if there are descriptors in the lines
        if "\n" in s:
            splitted_lines = s.split("\n")
            raw_results = [
                cls._try_get_multipath_descriptor(line.strip(), network) for line in splitted_lines
            ]
            # check that all entries return the same multipath descriptor
            # "None" entries are disallowed, to ensure that no deception is possible
            results = [r for r in raw_results if r]
            if results == raw_results:
                all_identical = all(
                    element.as_string_private() == results[0].as_string_private() for element in results
                )
                if all_identical:
                    return results[0]
                else:
                    # the string contins multiple inconsitent descriptors.
                    # Don't know how to handle this properly.
                    raise InconsistentDescriptors(f"The descriptors {splitted_lines} are inconsistent.")
            else:
                # if splitting lines didnt work, then remove the \n and try to recognize all as 1 descriptor
                s = s.replace("\n", "")

        try:
            multipath_descriptor = MultipathDescriptor.from_descriptor_str(s, network)
            if multipath_descriptor:
                logger.debug("detected descriptor")
                return multipath_descriptor
        except Exception:
            pass
        return None


class Data:
    """
    Recognized bitcoin data in a string, gives the data and the DataType
    """

    def __init__(self, data, data_type: DataType, network: bdk.Network) -> None:
        self.data = data
        self.data_type = data_type
        self.network = network

    def dump(self) -> Dict:
        return {"data": self.data_as_string(), "data_type": self.data_type.name}

    @classmethod
    def from_dump(cls, d: Dict, network: bdk.Network) -> "Data":
        data = cls.from_str(d["data"], network=network)
        d["data_type"] = DataType.from_name(d["data_type"])
        assert d["data_type"] == data.data_type
        return data

    def data_as_string(self) -> str:
        if not self.data:
            return str(self.data)
        if isinstance(self.data, str):
            return self.data
        if self.data_type == DataType.Bip21:
            return ConverterBip21(data=self.data).to_json()
        if self.data_type == DataType.Descriptor:
            return ConverterDescriptor(descriptor=self.data).to_str()
        if self.data_type == DataType.MultiPathDescriptor:
            return ConverterDescriptor(descriptor=self.data).to_str()
        if self.data_type == DataType.SignerInfo:
            return ConverterSignerInfo(info=self.data).to_json()
        if self.data_type == DataType.SignerInfos:
            return ConverterSignerInfos(infos=self.data, network=self.network).sparrow_format()
        if self.data_type == DataType.PSBT:
            return ConverterPSBT(psbt=self.data).to_str()
        if self.data_type == DataType.Tx:
            return ConverterTx(tx=self.data).to_str()
        if self.data_type == DataType.MultisigWalletExport and isinstance(
            self.data, ConverterMultisigWalletExport
        ):
            return self.data.to_str()

        return str(self.data)

    def __str__(self) -> str:
        return f"{self.data_type.name}: {self.data_as_string()}"

    @classmethod
    def from_binary(cls, raw: bytes, network: bdk.Network) -> "Data":
        # # Sequence of checks to identify the type of data in `s`
        # if decoded_bip21 := cls._try_decode_bip21(s, network=network):
        #     return Data(decoded_bip21, DataType.Bip21)

        # if is_xpub(s):
        #     data = convert_slip132_to_bip32(s) if is_slip132(s) else s
        #     return Data(data, DataType.Xpub)

        # if descriptor := cls._try_get_descriptor(s, network):
        #     return Data(descriptor, DataType.Descriptor)

        # if descriptor := cls._try_get_multipath_descriptor(s, network):
        #     return Data(descriptor, DataType.MultiPathDescriptor)

        # if is_valid_bitcoin_hash(s):
        #     return Data(s, DataType.Txid)

        # if is_valid_wallet_fingerprint(s):
        #     return Data(s, DataType.Fingerprint)

        if psbt := ConverterPSBT._try_decode_psbt_binary(raw):
            return Data(psbt, DataType.PSBT, network=network)

        if tx := ConverterTx._try_transaction_binary(raw):
            return Data(tx, DataType.Tx, network=network)

        # if signer_info := cls._try_extract_signer_info(s, network):
        #     return Data(signer_info, DataType.SignerInfo)

        # if signer_infos := cls._try_extract_signer_infos(s, network):
        #     return Data(signer_infos, DataType.SignerInfos)

        # if is_bip329(s):
        #     return Data(s, DataType.LabelsBip329)

        raise DecodingException(f"{raw} Could not be decoded with from_binary")  # type: ignore

    @classmethod
    def from_tx(cls, tx: bdk.Transaction, network: bdk.Network) -> "Data":
        assert isinstance(tx, bdk.Transaction)
        return Data(tx, DataType.Tx, network=network)

    @classmethod
    def from_psbt(cls, psbt: bdk.PartiallySignedTransaction, network: bdk.Network) -> "Data":
        assert isinstance(psbt, bdk.PartiallySignedTransaction)
        return Data(psbt, DataType.PSBT, network=network)

    @classmethod
    def from_descriptor(cls, descriptor: bdk.Descriptor, network: bdk.Network) -> "Data":
        assert isinstance(descriptor, bdk.Descriptor)
        return Data(descriptor, DataType.Descriptor, network=network)

    @classmethod
    def from_multipath_descriptor(
        cls, multipath_descriptor: MultipathDescriptor, network: bdk.Network
    ) -> "Data":
        assert isinstance(multipath_descriptor, MultipathDescriptor)
        return Data(multipath_descriptor, DataType.MultiPathDescriptor, network=network)

    @classmethod
    def from_str(cls, s: str, network: bdk.Network) -> "Data":
        s = s.strip()
        data = None

        # Sequence of checks to identify the type of data in `s`
        if decoded_bip21 := ConverterBip21._try_decode_bip21(s, network=network):
            return Data(decoded_bip21, DataType.Bip21, network=network)

        if ConverterXpub.is_xpub(s):
            data = ConverterXpub.normalized_to_bip32(s)
            return Data(data, DataType.Xpub, network=network)

        if descriptor := ConverterDescriptor._try_get_descriptor(s, network):
            return Data(descriptor, DataType.Descriptor, network=network)

        if descriptor := ConverterDescriptor._try_get_multipath_descriptor(s, network):
            return Data(descriptor, DataType.MultiPathDescriptor, network=network)

        if ConverterTxid.is_valid_bitcoin_hash(s):
            return Data(s, DataType.Txid, network=network)

        if ConverterFingerprint.is_valid_wallet_fingerprint(s):
            return Data(s, DataType.Fingerprint, network=network)

        if psbt := ConverterPSBT._try_decode_psbt(s):
            return Data(psbt, DataType.PSBT, network=network)

        if tx := ConverterTx._try_decode_serialized_transaction(s):
            return Data(tx, DataType.Tx, network=network)

        if signer_info := ConverterSignerInfo._try_extract_signer_info(s):
            return Data(signer_info, DataType.SignerInfo, network=network)

        if signer_infos := ConverterSignerInfos._try_extract_sparrow_signer_infos(s, network):
            return Data(signer_infos, DataType.SignerInfos, network=network)

        if signer_infos := ConverterSignerInfos._try_extract_multisig_signer_infos_coldcard_and_passport_qr(
            s, network
        ):
            return Data(signer_infos, DataType.SignerInfos, network=network)

        if ConverterBip329.is_bip329(s):
            return Data(s, DataType.LabelsBip329, network=network)

        if wallet_export := ConverterSignerInfos._try_multisig_wallet_export(s, network=network):
            # this is the Multisig wallet export of jade, and the signers (of course)  are different
            return Data(wallet_export, DataType.MultisigWalletExport, network=network)

        raise DecodingException(f"{s} Could not be decoded with from_str")

    def write_to_filedescriptor(self, fd):
        """

        This is useful for a temporary file
        fd, file_path = tempfile.mkstemp(
            suffix=f"_{key}.{self.file_extension}", prefix=""
        )

        a file descriptor for a normal file can be created with
                fd = os.open(filename, os.O_CREAT | os.O_WRONLY)
        """
        if self.data_type == DataType.Tx:
            with fdopen(fd, "wb") as file:
                file.write(bytes(self.data.serialize()))
        elif self.data_type == DataType.PSBT:
            with fdopen(fd, "wb") as file:
                file.write(base64.b64decode(self.data.serialize()))
        else:
            with fdopen(fd, "w") as file:
                file.write(self.data_as_string())



import logging
from typing import Any, List, Optional

import bdkpython as bdk
from base58 import b58decode_check
from hwilib.descriptor import MultisigDescriptor, PubkeyProvider, parse_descriptor
from hwilib.key import KeyOriginInfo, parse_path

from bitcoin_qr_tools.data import Data, SignerInfo
from bitcoin_qr_tools.urtypes.crypto import SCRIPT_EXPRESSION_TAG_MAP
from bitcoin_qr_tools.urtypes.crypto import CoinInfo as UR_CoinInfo
from bitcoin_qr_tools.urtypes.crypto import HDKey as UR_HDKey
from bitcoin_qr_tools.urtypes.crypto import Keypath as UR_KeyPath
from bitcoin_qr_tools.urtypes.crypto import MultiKey as UR_MultiKey
from bitcoin_qr_tools.urtypes.crypto import PathComponent as UR_PathComponent
from bitcoin_qr_tools.urtypes.crypto import ScriptExpression as UR_ScriptExpression
from bitcoin_qr_tools.utils import _flatten_descriptors

from .urtypes.crypto import Account as UR_ACCOUNT
from .urtypes.crypto import Output as UR_OUTPUT

logger = logging.getLogger(__name__)


class URComparator:
    @classmethod
    def _compare_path_component(cls, c1: UR_PathComponent, c2: UR_PathComponent, prefix: str = "") -> bool:
        all_same = True
        # Compare index
        if c1.index != c2.index:
            print(f"{prefix}DIFFERENCE in component index: {c1.index} != {c2.index}")
            all_same = False
        # else:
        #     print(f"{prefix}Same component index: {c1.index}")
        # Compare hardened
        if c1.hardened != c2.hardened:
            print(f"{prefix}DIFFERENCE in component hardened: {c1.hardened} != {c2.hardened}")
            all_same = False
        # else:
        #     print(f"{prefix}Same component hardened: {c1.hardened}")
        # Compare wildcard
        if c1.wildcard != c2.wildcard:
            print(f"{prefix}DIFFERENCE in component wildcard: {c1.wildcard} != {c2.wildcard}")
            all_same = False
        # else:
        #     print(f"{prefix}Same component wildcard: {c1.wildcard}")

        return all_same

    @classmethod
    def _compare_keypath(cls, kp1: Optional[UR_KeyPath], kp2: Optional[UR_KeyPath], prefix: str = "") -> bool:
        if kp1 is None and kp2 is None:
            # print(f"{prefix}Both Keypaths are None")
            return True
        elif kp1 is None:
            print(f"{prefix}DIFFERENCE: Keypath is None in first HDKey but not in second")
            return False
        elif kp2 is None:
            print(f"{prefix}DIFFERENCE: Keypath is None in second HDKey but not in first")
            return False

        all_same = True
        # Compare source_fingerprint
        if kp1.source_fingerprint != kp2.source_fingerprint:
            print(
                f"{prefix}DIFFERENCE in source_fingerprint: {kp1.source_fingerprint} != {kp2.source_fingerprint}"
            )
            all_same = False
        # else:
        # print(f"{prefix}Same source_fingerprint: {kp1.source_fingerprint}")
        # Compare depth
        if kp1.depth != kp2.depth:
            print(f"{prefix}DIFFERENCE in depth: {kp1.depth} != {kp2.depth}")
            all_same = False
        # else:
        # print(f"{prefix}Same depth: {kp1.depth}")
        # Compare components length
        if len(kp1.components) != len(kp2.components):
            print(
                f"{prefix}DIFFERENCE in number of components: {len(kp1.components)} != {len(kp2.components)}"
            )
            all_same = False
        else:
            # print(f"{prefix}Same number of components: {len(kp1.components)}")
            # Compare each component
            for i, (c1, c2) in enumerate(zip(kp1.components, kp2.components)):
                print(f"{prefix}Comparing component {i}:")
                if not cls._compare_path_component(c1, c2, prefix=prefix + "  "):
                    all_same = False

        return all_same

    @classmethod
    def _compare_use_info(cls, u1: Any, u2: Any, prefix: str = "") -> bool:
        if u1 is None and u2 is None:
            print(f"{prefix}Both use_info are None")
            return True
        elif u1 is None:
            print(f"{prefix}DIFFERENCE: use_info is None in first HDKey but not in second")
            return False
        elif u2 is None:
            print(f"{prefix}DIFFERENCE: use_info is None in second HDKey but not in first")
            return False

        all_same = True
        # Adjust fields according to your actual use_info structure
        if hasattr(u1, "network") and hasattr(u2, "network"):
            if u1.network != u2.network:
                print(f"{prefix}DIFFERENCE in use_info.network: {u1.network} != {u2.network}")
                all_same = False
            # else:
            #     print(f"{prefix}Same use_info.network: {u1.network}")
        else:
            print(f"{prefix}No 'network' attribute found on use_info. Adjust comparison as needed.")

        return all_same

    @classmethod
    def verbose_compare_hdkeys(cls, hdkey1: UR_HDKey, hdkey2: UR_HDKey) -> None:
        attributes = [
            "master",
            "key",
            "chain_code",
            "private_key",
            "parent_fingerprint",
            "name",
            "note",
        ]

        all_same = True
        # Compare simple attributes
        for attr in attributes:
            if not hasattr(hdkey1, attr) and not hasattr(hdkey2, attr):
                continue
            val1 = getattr(hdkey1, attr)
            val2 = getattr(hdkey2, attr)
            if val1 != val2:
                all_same = False
                print(f"DIFFERENCE in '{attr}':")
                print(f"  hdkey1.{attr} = {val1}")
                print(f"  hdkey2.{attr} = {val2}")
            # else:
            #     print(f"Same '{attr}': {val1}")

        # Compare use_info, origin, children using helpers
        print("\nComparing 'use_info':")
        if hasattr(hdkey1, "use_info") and hasattr(hdkey2, "use_info"):
            if not cls._compare_use_info(hdkey1.use_info, hdkey2.use_info, prefix="  "):
                all_same = False

        print("\nComparing 'origin':")
        if hasattr(hdkey1, "origin") and hasattr(hdkey2, "origin"):
            if not cls._compare_keypath(hdkey1.origin, hdkey2.origin, prefix="  "):
                all_same = False

        print("\nComparing 'children':")
        if hasattr(hdkey1, "children") and hasattr(hdkey2, "children"):
            if not cls._compare_keypath(hdkey1.children, hdkey2.children, prefix="  "):
                all_same = False

        if all_same:
            print("\nThe two UR_HDKey objects are identical.")
        else:
            print("\nDifferences found in the UR_HDKey objects above.")

    @classmethod
    def verbose_compare_output(cls, o1: UR_OUTPUT, o2: UR_OUTPUT) -> None:
        all_same = True

        # Compare script_expressions
        if len(o1.script_expressions) != len(o2.script_expressions):
            print(
                f"DIFFERENCE in number of script_expressions: {len(o1.script_expressions)} != {len(o2.script_expressions)}"
            )
            all_same = False
        else:
            # print(f"Same number of script_expressions: {len(o1.script_expressions)}")
            for i, (se1, se2) in enumerate(zip(o1.script_expressions, o2.script_expressions)):
                print(f"Comparing script_expressions[{i}]:")
                # Compare tag
                if se1.tag != se2.tag:
                    print(f"  DIFFERENCE in tag: {se1.tag} != {se2.tag}")
                    all_same = False
                # else:
                #     print(f"  Same tag: {se1.tag}")

                # Compare expression
                if se1.expression != se2.expression:
                    print(f"  DIFFERENCE in expression: {se1.expression} != {se2.expression}")
                    all_same = False
                # else:
                #     print(f"  Same expression: {se1.expression}")

        # Compare crypto_key
        print("\nComparing crypto_key:")
        if o1.crypto_key is None and o2.crypto_key is None:
            print("Both crypto_key are None")
        elif type(o1.crypto_key) != type(o2.crypto_key):
            print(f"DIFFERENCE: types {type(o1.crypto_key) } != { type(o2.crypto_key)}")
            all_same = False
        elif isinstance(o1.crypto_key, UR_HDKey) and isinstance(o2.crypto_key, UR_HDKey):
            cls.verbose_compare_hdkeys(o1.crypto_key, o2.crypto_key)
        elif isinstance(o1.crypto_key, UR_MultiKey) and isinstance(o2.crypto_key, UR_MultiKey):
            if len(o1.crypto_key.hd_keys) != len(o2.crypto_key.hd_keys):
                print(f"DIFFERENCE: .crypto_key.hd_keys have different length")
            for i, (k1, k2) in enumerate(zip(o1.crypto_key.hd_keys, o2.crypto_key.hd_keys)):
                print(f"Comparing {i}.th  keypair {k1, k2}")
                cls.verbose_compare_hdkeys(k1, k2)

        if all_same:
            print("\nThe two Output objects are identical.")
        else:
            print("\nDifferences found in the Output objects above.")

    @classmethod
    def verbose_compare_accounts(cls, a1: UR_ACCOUNT, a2: UR_ACCOUNT) -> None:
        if a1.master_fingerprint != a2.master_fingerprint:
            print(f"DIFFERENCE: master_fingerprint")

        for i, (o1, o2) in enumerate(zip(a1.output_descriptors, a2.output_descriptors)):
            print(f"Comapring {i}. output_descriptors\n")
            cls.verbose_compare_output(o1, o2)


# class URToolsUnsafe():


#     @classmethod
#     def  _hdkey_is_testnet(cls,  hdkey:UR_HDKey)->bool:
#         path_components = hdkey.origin.components
#         if len(path_components)>2 and path_components[1].index==1:
#             return True
#         return False


#     @classmethod
#     def  _hdkey_matches_network(cls,  hdkey:UR_HDKey, network:bdk.Network)->bool:
#         network_is_testnet            = network in [bdk.Network.REGTEST, bdk.Network.TESTNET, bdk.Network.SIGNET, ]
#         hdkey_is_testnet = cls._hdkey_is_testnet(hdkey)
#         return network_is_testnet == hdkey_is_testnet

#     @classmethod
#     def  _decode_hdkey(cls, hdkey:UR_HDKey, network:bdk.Network)->SignerInfo:
#         """
#         Unfortunately, coininfo.network / use-info.network is not set correctly in keystone
#         https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-007-hdkey.md#cddl-for-coin-info
#         so we use the https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-007-hdkey.md#cddl-for-key-path instead
#         """
#         xpub = hdkey.bip32_key(include_derivation_path=False)
#         fingerprint = binascii.hexlify(hdkey.origin.source_fingerprint).decode("utf-8")

#         path = hdkey.origin.path()
#         key_origin = f"m/{path}" if path  else "m"


#         if not cls._hdkey_matches_network(hdkey=hdkey, network=network) :
#             raise WrongNetwork(f"""Expected Network {network}. But {key_origin=} doesnt match.""")

#         return SignerInfo(fingerprint=fingerprint, key_origin=key_origin, xpub=xpub, name=hdkey.name)


#     @classmethod
#     def decode_output_as_signer_info(cls, output:UR_OUTPUT, network:bdk.Network)->SignerInfo:
#         hdkey = output.hd_key()
#         if not hdkey:
#             raise DecodingException(f"No HDKey available. Instead got {type(output.crypto_key)}")

#         data = cls._decode_hdkey(hdkey, network=network)
#         assert isinstance(data, SignerInfo)
#         data.name = f"p2{'-'.join([e.expression for e in output.script_expressions])}"
#         return data


class URTools:
    """
    See UR docs: https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-006-urtypes.md
    """

    @classmethod
    def decode_output_as_descriptor(cls, output: UR_OUTPUT, network: bdk.Network) -> Data:
        return Data.from_str(output.descriptor(), network=network)

    @classmethod
    def decode_account_as_signer_infos(cls, account: UR_ACCOUNT, network: bdk.Network) -> List[SignerInfo]:
        return [
            SignerInfo.decode_descriptor_as_signer_info(output.descriptor(), network=network)
            for output in account.output_descriptors
        ]

    @classmethod
    def _hd_key_to_pubkey_provider(cls, hdkey: UR_HDKey) -> PubkeyProvider:
        key_origin = KeyOriginInfo(
            fingerprint=hdkey.origin.source_fingerprint, path=parse_path(hdkey.origin.path())
        )
        xpub = hdkey.bip32_key(include_derivation_path=False)
        deriv_path = f"/{hdkey.children.path()}"
        return PubkeyProvider(origin=key_origin, pubkey=xpub, deriv_path=deriv_path)

    @classmethod
    def _key_origin_to_keypath(cls, key_origin_info: KeyOriginInfo, fingerprint: bytes = None) -> UR_KeyPath:
        """
        Convert a BIP32 path string into a Keypath object using hwi_parse_path.

        :param path_str: A string representing a BIP32 path (e.g. "m/84'/0'/0'")
        :param fingerprint: Optional 4-byte fingerprint to associate with the Keypath
        :return: A Keypath object populated with PathComponent objects.
        """
        # Use hwi_parse_path to get a list of uint32 integers representing the path
        parts = key_origin_info.path

        # Convert each index into a PathComponent
        components = []
        for part in parts:
            hardened = (part & 0x80000000) != 0
            index = part & ~0x80000000 if hardened else part
            components.append(UR_PathComponent(index=index, hardened=hardened))

        return UR_KeyPath(components=components, source_fingerprint=fingerprint, depth=len(components))

    @classmethod
    def _derivation_path_to_keypath(cls, derivation_path: str) -> UR_KeyPath:

        parts = derivation_path.lstrip("/").split("/")

        components = []
        for part in parts:
            if part == "*":
                # Wildcard component
                component = UR_PathComponent(index=None, hardened=False)
                component.wildcard = True
                components.append(component)
            else:
                # Regular BIP32 index
                integer_idx = int(part)
                hardened = (integer_idx & 0x80000000) != 0
                index = integer_idx & ~0x80000000 if hardened else integer_idx

                component = UR_PathComponent(index=index, hardened=hardened)
                component.wildcard = False
                components.append(component)

        depth = 0  # keystone lets this at 0.  len(components)
        return UR_KeyPath(components=components, source_fingerprint=None, depth=depth)

    @classmethod
    def _use_info(cls, is_testnet: bool) -> UR_CoinInfo:
        return UR_CoinInfo(type=0, network=int(is_testnet))

    @classmethod
    def _pubkey_provider_to_hdkey(cls, pubkey_provider: PubkeyProvider, name: str | None = None) -> UR_HDKey:
        # Decode the extended key (xpub / xprv)
        xpub = pubkey_provider.pubkey
        decoded = b58decode_check(xpub)

        # Extended key structure:
        # version (4 bytes) | depth (1 byte) | parent_fingerprint (4 bytes) |
        # child_number (4 bytes) | chain_code (32 bytes) | key_data (33 bytes)
        version = decoded[0:4]
        decoded[4]
        parent_fingerprint = decoded[5:9]
        child_number = decoded[9:13]
        chain_code = decoded[13:45]
        key_data = decoded[45:]

        # Determine if it's a private key
        if key_data[0] == 0x00:
            # Private key
            private_key = key_data[1:]
            key = private_key
        else:
            # Public key
            private_key = False
            key = key_data

        # Parse the origin path
        origin_keypath = cls._key_origin_to_keypath(
            pubkey_provider.origin, fingerprint=pubkey_provider.origin.fingerprint
        )

        # Parse the children path if provided
        children_keypath = (
            cls._derivation_path_to_keypath(pubkey_provider.deriv_path)
            if pubkey_provider.deriv_path
            else None
        )

        # Build props for UR_HDKey
        props = {
            "key": key,
            "chain_code": chain_code,
            "private_key": private_key,
            "use_info": cls._use_info(pubkey_provider.extkey.is_testnet),
            "origin": origin_keypath,
            "children": children_keypath,
            "parent_fingerprint": parent_fingerprint if parent_fingerprint != b"\x00\x00\x00\x00" else None,
            "name": name,
            "note": None,
        }

        return UR_HDKey(props)

    @classmethod
    def _script_expression_base(cls, name: str) -> UR_ScriptExpression:
        for tag in SCRIPT_EXPRESSION_TAG_MAP.values():
            if tag.expression == name:
                return tag

    @classmethod
    def _script_expressions(cls, name: str) -> List[UR_ScriptExpression]:
        expressions: List[UR_ScriptExpression] = []
        for single_expression in name.split("-"):
            for tag in SCRIPT_EXPRESSION_TAG_MAP.values():
                if tag.expression == single_expression:
                    expressions.append(tag)
        return expressions

    @classmethod
    def encode_ur_output(cls, descriptor_str: str, keystore_name: str | None = None) -> UR_OUTPUT:
        """_summary_

        Args:
            descriptor_str (str): _description_
            keystore_name (str): Name of the keystore.  Not supported for multisig.

        Raises:
            Exception: _description_
            NotImplementedError: _description_

        Returns:
            UR_OUTPUT: _description_
        """
        hwi_descriptor = parse_descriptor(desc=descriptor_str)

        if len(hwi_descriptor.pubkeys) == 1:
            descriptor_names, _ = SignerInfo._inner_most_pubkey_provider([hwi_descriptor])
            crypto_key = cls._pubkey_provider_to_hdkey(hwi_descriptor.pubkeys[0], name=keystore_name)
        elif len(hwi_descriptor.pubkeys) == 0 and hwi_descriptor.subdescriptors:
            # multisig
            flattened_descriptors = _flatten_descriptors(hwi_descriptor)
            descriptor_names = [d.name for d in flattened_descriptors]
            multisig_descriptor = flattened_descriptors[-1]
            if not isinstance(multisig_descriptor, MultisigDescriptor):
                raise Exception(f"descritpor not consistent with a multisig")

            hd_keys = [cls._pubkey_provider_to_hdkey(pubkey) for pubkey in multisig_descriptor.pubkeys]
            crypto_key = UR_MultiKey(threshold=multisig_descriptor.thresh, ec_keys=[], hd_keys=hd_keys)
        else:
            raise NotImplementedError(f"{len(hwi_descriptor.pubkeys)} pubkeys")

        script_expressions = [cls._script_expression_base(name) for name in descriptor_names]
        return UR_OUTPUT(script_expressions=script_expressions, crypto_key=crypto_key)

    @classmethod
    def encode_ur_account(cls, signer_infos: List[SignerInfo], descriptor_names: List[str]) -> UR_ACCOUNT:
        assert signer_infos, "Empty list signer_infos"
        first_info = signer_infos[0]

        output_descriptors = []
        for signer_info, descriptor_name in zip(signer_infos, descriptor_names):
            pubkey_provider = PubkeyProvider(
                origin=KeyOriginInfo.from_string(
                    signer_info.key_origin.replace("m", signer_info.fingerprint)
                ),
                pubkey=signer_info.xpub,
                deriv_path=signer_info.derivation_path,
            )
            crypto_key = cls._pubkey_provider_to_hdkey(pubkey_provider)

            script_expressions = cls._script_expressions(descriptor_name.replace("p2", ""))
            output = UR_OUTPUT(script_expressions=script_expressions, crypto_key=crypto_key)

            output_descriptors.append(output)

        return UR_ACCOUNT(
            master_fingerprint=bytes.fromhex(first_info.fingerprint), output_descriptors=output_descriptors
        )



import base64
import logging
from dataclasses import dataclass
from typing import List

import bdkpython as bdk

from bitcoin_qr_tools.bbqr.split import split_qrs
from bitcoin_qr_tools.data import Data, DataType
from bitcoin_qr_tools.ur_tools import URTools
from bitcoin_qr_tools.urtypes.bytes import BYTES
from bitcoin_qr_tools.urtypes.crypto.output import CRYPTO_OUTPUT
from bitcoin_qr_tools.urtypes.crypto.psbt import CRYPTO_PSBT
from bitcoin_qr_tools.utils import hex_to_serialized

from .ur.cbor_lite import CBOREncoder
from .ur.ur import UR
from .ur.ur_encoder import UREncoder

logger = logging.getLogger(__name__)


@dataclass
class QrExportType:
    name: str
    display_name: str


class QrExportTypes:
    bbqr = QrExportType("bbqr", "BBQr")
    ur = QrExportType("ur", "UR")
    text = QrExportType("text", "Text")

    @classmethod
    def as_list(cls) -> List[QrExportType]:
        return [
            export_type for name, export_type in cls.__dict__.items() if isinstance(export_type, QrExportType)
        ]


class UnifiedEncoder:
    "Create animated and static qr codes"

    @classmethod
    def bytes_to_ur_byte_fragments(cls, encoded: bytes, type="bytes", max_qr_size=50) -> List[str]:
        bcor_encoder = CBOREncoder()
        bcor_encoder.encodeBytes(encoded)
        ur = UR(type, bcor_encoder.get_bytes())

        encoder = UREncoder(ur, max_fragment_len=max_qr_size)
        fragments = []
        while not encoder.is_complete():
            part = encoder.next_part()
            fragments.append(part)
        return fragments

    @classmethod
    def string_to_ur_byte_fragments(cls, string_data: str, max_qr_size=50) -> List[str]:
        return cls.bytes_to_ur_byte_fragments(string_data.encode(), type="bytes", max_qr_size=max_qr_size)

    @staticmethod
    def _max_qr_code_version(num_modules: int) -> int:
        """
        Determine the maximum QR code version based on the number of modules per side,
        returning the next lowest version if an exact match isn't found.

        This function adjusts for inputs that do not directly correspond to a specific QR code version,
        opting instead to return the highest version whose module count is less than or equal to the given number.

        The formula used is: version = (num_modules - 21) // 4 + 1

        Parameters:
        - num_modules (int): The number of modules per side of the QR code.

        Returns:
        - int: The maximum version of a QR code for which the module count per side is less than or equal to the provided number.

        The complete QR code version-to-modules mapping is:
        - Version 1: 21x21 modules
        - Version 2: 25x25 modules
        - Version 3: 29x29 modules
        - Version 4: 33x33 modules
        - Version 5: 37x37 modules
        - Version 6: 41x41 modules
        - Version 7: 45x45 modules
        - Version 8: 49x49 modules
        - Version 9: 53x53 modules
        - Version 10: 57x57 modules
        - Version 11: 61x61 modules
        - Version 12: 65x65 modules
        - Version 13: 69x69 modules
        - Version 14: 73x73 modules
        - Version 15: 77x77 modules
        - Version 16: 81x81 modules
        - Version 17: 85x85 modules
        - Version 18: 89x89 modules
        - Version 19: 93x93 modules
        - Version 20: 97x97 modules
        - Version 21: 101x101 modules
        - Version 22: 105x105 modules
        - Version 23: 109x109 modules
        - Version 24: 113x113 modules
        - Version 25: 117x117 modules
        - Version 26: 121x121 modules
        - Version 27: 125x125 modules
        - Version 28: 129x129 modules
        - Version 29: 133x133 modules
        - Version 30: 137x137 modules
        - Version 31: 141x141 modules
        - Version 32: 145x145 modules
        - Version 33: 149x149 modules
        - Version 34: 153x153 modules
        - Version 35: 157x157 modules
        - Version 36: 161x161 modules
        - Version 37: 165x165 modules
        - Version 38: 169x169 modules
        - Version 39: 173x173 modules
        - Version 40: 177x177 modules

        Example:
        - A QR code with 26 modules per side corresponds to Version 2 since it's the highest version not exceeding 26 modules.
        """
        if num_modules < 21:
            return 1

        # Calculate version based on the input module size.
        version = (num_modules - 21) // 4 + 1

        # Ensure the calculated version does not exceed the bounds of the QR specification (1 to 40).
        if version < 1:
            return 1
        elif version > 40:
            version = 40  # Cap the version at 40 if it exceeds the maximum.

        # Check if calculated version really matches the max version not exceeding the given module size.
        # This adjusts down if the input size does not perfectly align with the calculated version's expected size.
        if (21 + 4 * (version - 1)) > num_modules:
            version -= 1

        return version

    @classmethod
    def generate_fragments_for_qr(cls, data: Data, qr_export_type: QrExportType, max_qr_size=50) -> List[str]:
        if qr_export_type.name == QrExportTypes.text.name:
            return [data.data_as_string()]
        elif qr_export_type.name == QrExportTypes.ur.name:
            serialized = data.data_as_string()

            if len(serialized) <= max_qr_size:
                return [serialized]

            if data.data_type == DataType.Tx:
                return cls.bytes_to_ur_byte_fragments(
                    hex_to_serialized(serialized), type=BYTES.type, max_qr_size=max_qr_size
                )
            elif data.data_type == DataType.PSBT:
                return cls.bytes_to_ur_byte_fragments(
                    base64.b64decode(serialized.encode()), type=CRYPTO_PSBT.type, max_qr_size=max_qr_size
                )
            elif data.data_type == DataType.Descriptor:
                return cls.bytes_to_ur_byte_fragments(
                    URTools.encode_ur_output(descriptor_str=serialized).to_cbor(),
                    type=CRYPTO_OUTPUT.type,
                    max_qr_size=max_qr_size,
                )
            else:
                return cls.bytes_to_ur_byte_fragments(
                    serialized.encode(), type="bytes", max_qr_size=max_qr_size
                )

        elif qr_export_type.name == QrExportTypes.bbqr.name:
            if data.data_type == DataType.Tx:
                file_type = "T"
                assert isinstance(data.data, bdk.Transaction)
                raw = bytes(data.data.serialize())
            elif data.data_type == DataType.PSBT:
                file_type = "P"
                assert isinstance(data.data, bdk.PartiallySignedTransaction)
                raw = base64.b64decode(data.data.serialize())
            else:
                file_type = "U"
                raw = data.data_as_string().encode()

            version, parts = split_qrs(
                raw, file_type, max_version=cls._max_qr_code_version(num_modules=max_qr_size)
            )
            return parts

        raise Exception(f"Unknown qr_type {qr_export_type}")



import sys

import cv2
import pygame
from PyQt6.QtCore import Qt, QTimer
from PyQt6.QtGui import QImage, QPixmap
from PyQt6.QtWidgets import QApplication, QLabel, QVBoxLayout, QWidget


class CV2Camera:
    def __init__(self, device_index=0, resolution=None, mode="RGB"):
        self.device_index = device_index
        self.resolution = resolution
        self._cam = None
        self._open = False
        self.mode = mode

    def start(self):
        self._cam = cv2.VideoCapture(self.device_index)

        if not self._cam.isOpened():
            raise ValueError("Could not open camera")
        self._open = True

    def stop(self):
        if self._open and self._cam:
            self._cam.release()
            self._open = False

    def get_image(self) -> pygame.Surface:
        if not self._open or not self._cam:
            raise ValueError("Camera must be started")
        ret, image = self._cam.read()
        if not ret:
            raise ValueError("Failed to capture image")
        if self.mode == "RGB":
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Convert numpy array to pygame.Surface
        surface = pygame.surfarray.make_surface(image.swapaxes(0, 1))
        return surface

    def __del__(self):
        self.stop()


if __name__ == "__main__":

    class MainWindow(QWidget):
        def __init__(self):
            super().__init__()
            self.camera = CV2Camera(device_index=0, resolution=(640, 480), mode="RGB")
            self.camera.start()

            self.label_image = QLabel(self)
            self._layout = QVBoxLayout(self)
            self._layout.addWidget(self.label_image)
            self.timer = QTimer(self)
            self.timer.timeout.connect(self.update_frame)
            self.timer.start(30)  # Update at ~33fps

        def showSurface(self, surface: pygame.Surface, scale_to=(640, 480)):

            array3d = pygame.surfarray.array3d(surface)
            height, width, _ = array3d.shape
            bytes_per_line = 3 * width
            q_image = QImage(array3d.data, width, height, bytes_per_line, QImage.Format.Format_RGB888)

            # Convert QImage to QPixmap
            pixmap = QPixmap.fromImage(q_image)

            # Scale the pixmap if scaling dimensions are provided
            if scale_to:
                pixmap = pixmap.scaled(
                    scale_to[0],
                    scale_to[1],
                    Qt.AspectRatioMode.KeepAspectRatio,
                    Qt.TransformationMode.SmoothTransformation,
                )

            self.label_image.setPixmap(pixmap)

        def update_frame(self):
            frame = self.camera.get_image()
            self.showSurface(frame)
            # image = QImage(frame.data, frame.shape[1], frame.shape[0], QImage.Format.Format_RGB888)
            # pixmap = QPixmap.fromImage(image)
            # self.label_image.setPixmap(pixmap)

        def closeEvent(self, event):
            self.timer.stop()
            self.camera.stop()
            super().closeEvent(event)

    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())



import logging
from collections import deque
from concurrent.futures import ThreadPoolExecutor

from bitcoin_qr_tools.i18n import translate
from bitcoin_qr_tools.rtsp_camera import RTSPCamera
from bitcoin_qr_tools.screen_camera import ScreenCamera

logger = logging.getLogger(__name__)

import signal
import sys
import time
from typing import Any, List, Tuple, Union

import cv2
import numpy as np
import pygame
import pygame.camera
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QAction
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QMenu,
    QMessageBox,
    QSizePolicy,
    QSlider,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from ..cv2camera import CV2Camera


def threaded_list(f, args):
    with ThreadPoolExecutor() as executor:
        res = [executor.submit(f, arg) for arg in args]
    return [r.result() for r in res]


class BarcodeData:
    def __init__(self, data, rect):
        self.data = data  # The decoded text of the barcode
        self.rect = rect  # The bounding rectangle of the barcode as a tuple (x, y, width, height)

    def __repr__(self):
        return f"BarcodeData(data={self.data}, rect={self.rect})"


TypeSomeCamera = Union[CV2Camera, RTSPCamera, pygame.camera.Camera, ScreenCamera]


class VideoWidget(QWidget):
    signal_raw_qr_data = pyqtSignal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(translate("video", "Camera"))
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.WindowStaysOnTopHint)
        self.frame_counter: deque[float] = deque(maxlen=5000)
        self.last_display_fps = time.time()
        self.cv2 = None
        self.pyzbar = None
        try:
            # check if loading works
            # it depend on zlib installed in the os
            from pyzbar import pyzbar

            self.pyzbar = pyzbar
            logger.info("Load pyzbar successful")
        except:
            logger.info("Could not load pyzbar. Trying to load fallback cv2")
            import cv2

            self.cv2 = cv2

        self.label_image = QLabel()
        self.lower_widget = QWidget()
        self.lower_widget_layout = QHBoxLayout(self.lower_widget)

        self.combo_cameras = QComboBox()
        self.combo_cameras.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        self.label_camera_choose = QLabel(translate("video", "Camera:"))
        self.lower_widget_layout.addWidget(self.label_camera_choose)
        self.lower_widget_layout.addWidget(self.combo_cameras)

        # Create the tool button
        settingsButton = QToolButton(self)
        settingsButton.setText(translate("video", "Settings"))
        settingsButton.setToolButtonStyle(
            Qt.ToolButtonStyle.ToolButtonTextBesideIcon
        )  # Show text beside the icon
        self.lower_widget_layout.addWidget(settingsButton)

        # Create postprocess_checkbox
        self.preset_process_checkbox = QCheckBox(
            translate("video", "Enhance picture for detection"), parent=self
        )
        self.preset_process_checkbox.stateChanged.connect(self.on_preset_process_checkbox)
        self.preset_process_checkbox.setChecked(False)

        # middle widget
        self.middle_widget = QWidget()
        self.middle_widget.setVisible(False)
        self.middle_widget_layout = QHBoxLayout(self.middle_widget)

        # Create zoom slider
        self.zoom_slider = QSlider(Qt.Orientation.Horizontal, self)
        self.zoom_slider.setMinimum(10)
        self.zoom_slider.setMaximum(50)
        self.zoom_slider.setValue(10)  # Default value
        self.zoom_slider.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.zoom_slider.setTickInterval(1)

        self.label_zoom_slider = QLabel(translate("video", "Zoom:"))
        self.middle_widget_layout.addWidget(self.label_zoom_slider)
        self.middle_widget_layout.addWidget(self.zoom_slider)

        # Create brightness slider
        self.brightness_slider = QSlider(Qt.Orientation.Horizontal, self)
        self.brightness_slider.setMinimum(0)
        self.brightness_slider.setMaximum(200)
        self.brightness_slider.setValue(100)
        self.brightness_slider.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.brightness_slider.setTickInterval(10)
        self.brightness_slider.valueChanged.connect(self.on_change_brightness)

        self.label_brightness_slider = QLabel(translate("video", "Brightness (reduce for bright displays):"))
        self.middle_widget_layout.addWidget(self.label_brightness_slider)
        self.middle_widget_layout.addWidget(self.brightness_slider)

        # Create postprocess_checkbox
        self.post_process_checkbox = QCheckBox(parent=self)
        self.post_process_checkbox.setChecked(False)
        self.label_post_process_checkbox = QLabel(translate("video", "Postprocess"))
        self.middle_widget_layout.addWidget(self.label_post_process_checkbox)
        self.middle_widget_layout.addWidget(self.post_process_checkbox)

        # Create a menu for the button
        menu = QMenu("", self)
        settingsButton.setMenu(menu)
        settingsButton.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        action_show_camera_controls = QAction(translate("video", "Show camera controls"), self)
        menu.addAction(action_show_camera_controls)
        action_show_camera_controls.setCheckable(True)
        action_show_camera_controls.triggered.connect(self.middle_widget.setVisible)

        action_add_rtsp_camera = QAction(translate("video", "Add RTSP Camera"), self)
        menu.addAction(action_add_rtsp_camera)
        action_add_rtsp_camera.triggered.connect(self.prompt_rtsp_url)

        pygame.camera.init()
        for index, camera_name, camera in self.get_valid_cameras():
            self.combo_cameras.addItem(str(camera_name), userData=camera)

        layout = QVBoxLayout()
        layout.addWidget(self.label_image)
        layout.addWidget(self.preset_process_checkbox)
        layout.addWidget(self.middle_widget)
        layout.addWidget(self.lower_widget)

        self.setLayout(layout)

        self.current_camera: TypeSomeCamera | None = None

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_frame)
        self.timer.start(int(1000.0 / 30.0))  # 30 FPS

        # Add "Screen" option for screen capture
        self.combo_cameras.addItem(translate("video", "Screen"), userData=ScreenCamera())

        # switch to the last camera that isn
        if self.combo_cameras.count():
            last_camera_idx = max(self.combo_cameras.count() - 2, 0)
            self.combo_cameras.setCurrentIndex(last_camera_idx)
            self.switch_camera(last_camera_idx)

        # signals
        self.combo_cameras.currentIndexChanged.connect(self.switch_camera)

        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)  # Handle Ctrl+C

    def signal_handler(self, sig, frame):
        logger.debug("Signal received:", sig)
        self.stop_timer_and_stop_all_cameras()
        sys.exit(0)

    @property
    def zoom(self) -> float:
        return self.zoom_slider.value() / 10

    def prompt_rtsp_url(self):
        text, ok = QInputDialog.getText(
            self, translate("video", "Enter RTSP URL"), translate("video", "RTSP URL:"), text="rtsp://"
        )
        if ok and text:
            self.add_rtsp_camera(text)

    def add_rtsp_camera(self, url: str):
        # Here you would add your actual logic to handle the RTSP camera
        # This is a placeholder function to illustrate what you might do
        logger.debug("Adding RTSP Camera with URL:", url)
        # Assume validation of the URL or further actions here

        for enable_udp in [False, True]:
            temp_camera = RTSPCamera(url)
            try:
                temp_camera.start(enable_udp=enable_udp)
                temp_camera.stop()
                self.combo_cameras.addItem(str(url), userData=temp_camera)
                self.combo_cameras.setCurrentIndex(self.combo_cameras.count() - 1)
                self.switch_camera(self.combo_cameras.count() - 1)
                return
            except:
                logger.debug(f"{temp_camera} with enable_udp {enable_udp} could not be opened")

        QMessageBox().warning(
            None, translate("video", "Error"), translate("video", "The camera could not be opened")
        )

    def stop_timer_and_stop_all_cameras(self):
        self.timer.stop()

        if self.current_camera:
            self.set_brightness_defaults(self.current_camera)

        for i in range(self.combo_cameras.count()):
            camera: TypeSomeCamera = self.combo_cameras.itemData(i)
            if camera:
                try:
                    logger.debug(f"Stopping {camera}")
                    camera.stop()
                except SystemError as e:
                    if "ioctl(VIDIOC_STREAMOFF) failure" in str(e):
                        pass  # this is expected
                    else:
                        logger.debug(str(e))
                except Exception as e:
                    # if it is already closed, it gives an exception.
                    logger.debug(str(e))

    def closeEvent(self, event: QtGui.QCloseEvent | None) -> None:
        self.stop_timer_and_stop_all_cameras()

        # If you call the parent's closeEvent(), it will proceed to close the widget
        super().closeEvent(event)

    def set_brightness(self, value=128, camera=None):
        selected_camera = self.combo_cameras.currentData()
        if isinstance(selected_camera, (pygame.camera.Camera, CV2Camera)):
            self.change_brightness(selected_camera, value)

    def set_brightness_defaults(self, camera: Any):
        self.set_brightness(128, camera=camera)

    def on_change_brightness(self, value: int):
        selected_camera = self.combo_cameras.currentData()
        if isinstance(selected_camera, (pygame.camera.Camera, CV2Camera)):
            self.change_brightness(selected_camera, value)

    def get_brightness(self, camera: Any) -> Tuple[float, float, float]:
        if isinstance(camera, pygame.camera.Camera):
            _flipx, _flipy, brightness = camera.get_controls()
            return 0, 255, brightness

        elif isinstance(camera, CV2Camera) and camera._cam:
            brightness = camera._cam.get(cv2.CAP_PROP_BRIGHTNESS)
            return 0, 255, brightness

        return 0, 255, 128

    def change_brightness(self, camera: Union[pygame.camera.Camera, CV2Camera], value: float):
        v = cv2.CAP_PROP_BRIGHTNESS

        target_value = int(value)

        if isinstance(camera, pygame.camera.Camera):
            _flipx, _flipy, old_value = camera.get_controls()
            camera.set_controls(False, False, target_value)
            _flipx, _flipy, new_value = camera.get_controls()

            if new_value != old_value:
                logger.debug(f"Changed {v} from {old_value} --> { new_value }")

        elif isinstance(camera, CV2Camera) and camera._cam:
            old_value = camera._cam.get(v)
            camera._cam.set(v, target_value)
            new_value = camera._cam.get(v)

            if new_value != old_value:
                logger.debug(f"Changed {v} from {old_value} --> { new_value }")

    def on_preset_process_checkbox(self, state: Any):
        self.post_process_checkbox.setChecked(self.preset_process_checkbox.isChecked())
        if self.preset_process_checkbox.isChecked():
            self.zoom_slider.setValue(25)
            self.brightness_slider.setValue(20)
        else:
            self.zoom_slider.setValue(0)
            self.brightness_slider.setValue(128)

    def find_best_resolution(self, camera_index=0):
        # Common resolutions to try
        common_resolutions = [(1280, 720), (1024, 768), (800, 600), (640, 480)]

        camera = cv2.VideoCapture(camera_index)
        best_resolution = (0, 0)

        for resolution in common_resolutions:
            # Set camera resolution
            camera.set(cv2.CAP_PROP_FRAME_WIDTH, resolution[0])
            camera.set(cv2.CAP_PROP_FRAME_HEIGHT, resolution[1])

            # Check if the camera accepts this resolution
            test_width = camera.get(cv2.CAP_PROP_FRAME_WIDTH)
            test_height = camera.get(cv2.CAP_PROP_FRAME_HEIGHT)

            if (test_width, test_height) == resolution:
                best_resolution = resolution
                break

        camera.release()
        return best_resolution

    def get_pygame_camera(self, camera_name: str | int, index) -> pygame.camera.Camera | None:
        try:
            resolution = self.find_best_resolution(index)
            temp_camera = pygame.camera.Camera(camera_name, resolution)
            temp_camera.start()
            temp_camera.stop()
            logger.debug(f"Found pygame.camera.Camera({camera_name}, {resolution})")
            return temp_camera
        except Exception as e:
            logger.debug(f"Could not get  pygame.camera.Camera({camera_name}). {e} ")
        return None

    def get_cv2camera(self, index: int) -> CV2Camera | None:
        try:
            resolution = self.find_best_resolution(index)
            temp_camera = CV2Camera(index, resolution)
            temp_camera.start()
            temp_camera.stop()
            logger.debug(f"Found CV2Camera({index}, {resolution})")
            return temp_camera
        except Exception as e:
            logger.debug(f"Could not get  CV2Camera({index}). {e} ")

        return None

    def get_valid_cameras(self) -> List[Tuple[int, str, TypeSomeCamera]]:
        valid_cameras: List[Tuple[int, str, TypeSomeCamera]] = []
        for index, camera_name in enumerate(pygame.camera.list_cameras()):
            temp_camera = self.get_pygame_camera(camera_name, index) or self.get_cv2camera(index)
            if temp_camera:
                valid_cameras.append((index, camera_name, temp_camera))

        if not valid_cameras:
            temp_camera = self.get_pygame_camera(0, 0) or self.get_cv2camera(0)
            if temp_camera:
                valid_cameras.append((0, str(0), temp_camera))

        return valid_cameras

    def switch_camera(self, index: int):
        selected_camera: TypeSomeCamera = self.combo_cameras.currentData()

        self.set_brightness_defaults(self.current_camera)
        self.on_preset_process_checkbox(False)
        if self.current_camera:
            try:
                self.current_camera.stop()
            except:
                pass

        self.current_camera = selected_camera

        if self.current_camera:
            self.current_camera.start()
            time.sleep(0.1)  # Add a short delay
        else:
            # The logic of capture is in update_frame
            pass

        b_min, b_max, b = self.get_brightness(self.current_camera)
        self.brightness_slider.setMinimum(int(b_min))
        self.brightness_slider.setMaximum(int(b_max))
        self.brightness_slider.setValue(int(b))
        self.on_preset_process_checkbox(self.preset_process_checkbox.isChecked())

    def _numpy_to_surface(self, numpy_image: np.ndarray) -> pygame.Surface:
        # Convert numpy image (RGB) to pygame surface
        return pygame.surfarray.make_surface(numpy_image.transpose((1, 0, 2)))

    def _on_draw_surface(self, surface, barcode: BarcodeData):
        x, y, w, h = barcode.rect
        pygame.draw.rect(surface, (0, 255, 0), (x, y, w, h), 2)

    @staticmethod
    def crop(
        image_surface: pygame.Surface, left: float, right: float, top: float, bottom: float
    ) -> pygame.Surface:
        """
        Crop the image based on the specified proportions from each edge using Pygame.

        :param image_surface: The full image as a pygame.Surface.
        :param left: Proportion of the width to crop from the left.
        :param right: Proportion of the width to crop from the right.
        :param top: Proportion of the height to crop from the top.
        :param bottom: Proportion of the height to crop from the bottom.
        :return: Cropped image as a pygame.Surface.
        """
        w, h = image_surface.get_size()
        left_idx = int(w * left)
        right_idx = int(w * (1 - right))
        top_idx = int(h * top)
        bottom_idx = int(h * (1 - bottom))

        # Ensure the indices are within the image dimensions
        left_idx = max(left_idx, 0)
        right_idx = min(right_idx, w)
        top_idx = max(top_idx, 0)
        bottom_idx = min(bottom_idx, h)

        # Define the rectangle for the subsurface
        rect = pygame.Rect(left_idx, top_idx, right_idx - left_idx, bottom_idx - top_idx)

        # Return the subsurface
        return image_surface.subsurface(rect)

    def get_barcodes(self, array: np.ndarray) -> List[BarcodeData]:
        if self.pyzbar:
            decoded_codes = self.pyzbar.decode(array)
            return [BarcodeData(data=decoded.data, rect=decoded.rect) for decoded in decoded_codes]
        elif self.cv2:
            array = self.cv2.transpose(array)
            array = self.cv2.cvtColor(array, self.cv2.COLOR_RGB2BGR)
            # Use OpenCV's QRCodeDetector to detect and decode the QR code.
            detector = self.cv2.QRCodeDetector()
            val, points, straight_qrcode = detector.detectAndDecode(array)

            barcodes = []
            if val:
                # If a QR code is detected, 'val' contains the decoded text.
                # 'points' contain the coordinates of the QR code corners.
                if points is not None:
                    # Convert points to a more manageable format
                    points = points[0]  # Points are returned in a nested array.

                    # Calculate the bounding rectangle using the points.
                    y, x, w, h = self.cv2.boundingRect(points.astype(np.int32))
                    rect = (x, y, w, h)  # Store the rectangle as a tuple

                    barcode = BarcodeData(val.encode(), rect)
                    barcodes.append(barcode)

            return barcodes
        else:
            return []

    @staticmethod
    def preprocess_for_qr_detection(
        array: np.ndarray,
        gauss_kernel_size: int = 5,
        single_channel: int | None = None,
        threshold_blockSize: int = 11,
        threshold_C: float = 0,
    ):
        """
        Preprocess a 3D NumPy array for QR detection, returning a 3D NumPy array.

        Args:
            array (np.ndarray): Input 3D NumPy array in RGB format.

        Returns:
            np.ndarray: Processed 3D NumPy array in RGB format.
        """
        # Ensure the array is in the right format (RGB)
        if array.ndim != 3 or array.shape[2] != 3:
            raise ValueError("Input array must be a 3D NumPy array with 3 channels (RGB)")

        # Convert RGB to Grayscale
        if single_channel is None:
            gray = cv2.cvtColor(array, cv2.COLOR_RGB2GRAY)
        else:
            gray = array[:, :, single_channel]

        # logger.debug("Average brightness (grayscale):",  np.mean(gray))

        # Enhance contrast using histogram equalization
        gray = cv2.equalizeHist(gray)

        # # Apply Gaussian blur to reduce noise
        gray = cv2.GaussianBlur(gray, (gauss_kernel_size, gauss_kernel_size), 0)

        # # Apply adaptive thresholding
        gray = cv2.adaptiveThreshold(
            gray, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, threshold_blockSize, threshold_C
        )

        return cv2.cvtColor(gray, cv2.COLOR_GRAY2RGB)

    def update_frame(self):
        self.frame_counter.append(time.time())
        if time.time() - self.last_display_fps > 10:
            self.last_display_fps = time.time()
            if self.frame_counter[-1] - self.frame_counter[0] > 0:
                logger.debug(
                    f"fps: {len(self.frame_counter)/(self.frame_counter[-1]- self.frame_counter[0] )  }"
                )

        barcodes: List[BarcodeData] = []
        if self.current_camera:
            try:
                surface = self.current_camera.get_image()
            except:
                # logger.debug("Could not get image")
                return
        else:
            return

        crop_value = min((1 - 1 / self.zoom) / 2, 0.48)
        surface = self.crop(
            surface,
            crop_value,
            crop_value,
            crop_value,
            crop_value,
        )
        array = array_original = pygame.surfarray.array3d(surface)
        if self.post_process_checkbox.isChecked():
            array = self.preprocess_for_qr_detection(array_original.copy())
            surface = pygame.surfarray.make_surface(array)
        surface = pygame.transform.flip(surface, False, True)
        surface = pygame.transform.rotate(surface, -90)

        def transform_and_detect(values: Tuple[int, int] | None) -> List[BarcodeData]:
            if values is None:
                return self.get_barcodes(array_original)
            gauss_kernel_size, thres = values
            try:
                return self.get_barcodes(
                    self.preprocess_for_qr_detection(
                        array_original.copy(), gauss_kernel_size=gauss_kernel_size, threshold_blockSize=thres
                    )
                )
            except:
                return []

        arguments = [None, (5, 11), (5, 21), (3, 11), (9, 31), (11, 35), (7, 21)]
        list_of_barcodes = threaded_list(transform_and_detect, arguments)
        barcodes = sum(list_of_barcodes, [])
        if barcodes:
            logger.debug(
                f"Found barcodes with parameters {[argument for barcodes, argument in  zip(list_of_barcodes, arguments) if barcodes]}"
            )

        sorted_barcodes = sorted(barcodes, key=lambda item: len(item.data), reverse=True)
        # only show the 1. barcode (with the longest data)
        selected_barcode = sorted_barcodes[0] if sorted_barcodes else None
        if selected_barcode and not selected_barcode.data:
            selected_barcode = None

        if selected_barcode:
            self._on_draw_surface(surface, selected_barcode)

        surface = pygame.transform.flip(surface, False, True)
        self.showSurface(surface, scale_to=(640, 480))

        if selected_barcode:
            self.signal_raw_qr_data.emit(selected_barcode.data)

    def showSurface(self, surface: pygame.Surface, scale_to=(640, 480)):

        array3d = pygame.surfarray.array3d(surface)
        height, width, _ = array3d.shape
        bytes_per_line = 3 * width
        q_image = QtGui.QImage(array3d.data, width, height, bytes_per_line, QtGui.QImage.Format.Format_RGB888)

        # Convert QImage to QPixmap
        pixmap = QtGui.QPixmap.fromImage(q_image)

        # Scale the pixmap if scaling dimensions are provided
        if scale_to:
            pixmap = pixmap.scaled(
                scale_to[0],
                scale_to[1],
                QtCore.Qt.AspectRatioMode.KeepAspectRatio,
                QtCore.Qt.TransformationMode.SmoothTransformation,
            )

        self.label_image.setPixmap(pixmap)


class DemoVideoWidget(VideoWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.label_qr = QtWidgets.QTextEdit()

        self.layout().addWidget(self.label_qr)  # type: ignore

        self.signal_raw_qr_data.connect(self.show_qr)

    def show_qr(self, qr_data):
        s = qr_data.decode("utf-8")
        logger.debug(s)
        self.label_qr.setText(s)



import logging
import os
import sys

import cv2
import pygame
from PyQt6.QtCore import Qt, QTimer
from PyQt6.QtGui import QImage, QPixmap
from PyQt6.QtWidgets import QApplication, QLabel, QVBoxLayout, QWidget

logger = logging.getLogger(__name__)


class RTSPCamera:
    def __init__(self, rtsp_url: str):
        """
        Initialize the RTSP camera with the given URL and resolution.

        :param rtsp_url: str, the RTSP URL to the camera stream
        :param resolution: tuple, the resolution of the camera as (width, height)
        """
        self.rtsp_url = rtsp_url
        self._cam: cv2.VideoCapture | None = None
        self._open = False

    def start(self, enable_udp=False):
        """
        Start the RTSP camera capture.
        """

        if enable_udp:
            os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = "rtsp_transport;udp"

        self._cam = cv2.VideoCapture(self.rtsp_url)

        self.width = int(self._cam.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.height = int(self._cam.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.fps = self._cam.get(cv2.CAP_PROP_FPS)  # Reading the frames per second

        logger.debug(f"Stream resolution: {self.width}x{self.height}, FPS: {self.fps}")

        if not self._cam.isOpened():
            raise ValueError("Could not open camera stream.")

        self._open = True

    def stop(self):
        """
        Stop the RTSP camera capture.
        """
        if self._open and self._cam:
            self._cam.release()
            self._open = False

    def get_image(self):
        """
        Capture a single frame from the RTSP stream.

        :return: The captured frame as a numpy array, or None if capture failed.
        """
        if not self._open or not self._cam:
            raise RuntimeError("Camera is not initialized, call start() first")

        ret, frame = self._cam.read()
        if not ret:
            raise RuntimeError("Failed to capture image")

        # Convert numpy array to pygame.Surface
        surface = pygame.surfarray.make_surface(frame.swapaxes(0, 1))
        surface = pygame.transform.flip(surface, True, False)
        return surface

    def __del__(self):
        """
        Ensure the camera stream is released when the object is destroyed.
        """
        self.stop()


if __name__ == "__main__":

    class MainWindow(QWidget):
        def __init__(self, rtsp_url):
            super().__init__()
            self.camera = RTSPCamera(rtsp_url)
            self.camera.start(enable_udp=True)

            self.label_image = QLabel(self)
            self._layout = QVBoxLayout(self)
            self._layout.addWidget(self.label_image)
            self.timer = QTimer(self)
            self.timer.timeout.connect(self.update_frame)
            self.timer.start(30)  # Update at ~33fps

        def showSurface(self, surface: pygame.Surface, scale_to=(640, 480)):
            array3d = pygame.surfarray.array3d(surface)
            height, width, _ = array3d.shape
            bytes_per_line = 3 * width
            q_image = QImage(array3d.data, width, height, bytes_per_line, QImage.Format.Format_RGB888)

            # Convert QImage to QPixmap
            pixmap = QPixmap.fromImage(q_image)

            # Scale the pixmap if scaling dimensions are provided
            if scale_to:
                pixmap = pixmap.scaled(
                    scale_to[0],
                    scale_to[1],
                    Qt.AspectRatioMode.KeepAspectRatio,
                    Qt.TransformationMode.SmoothTransformation,
                )

            self.label_image.setPixmap(pixmap)

        def update_frame(self):
            frame = self.camera.get_image()

            self.showSurface(frame)
            # image = QImage(frame.data, frame.shape[1], frame.shape[0], QImage.Format.Format_RGB888)
            # pixmap = QPixmap.fromImage(image)
            # self.label_image.setPixmap(pixmap)

        def closeEvent(self, event):
            self.timer.stop()
            self.camera.stop()
            super().closeEvent(event)

    app = QApplication(sys.argv)
    window = MainWindow("rtsp://192.168.178.27:8086/webcam")
    window.show()
    sys.exit(app.exec())



import logging
from typing import List

from hwilib.descriptor import Descriptor

BITCOIN_BIP21_URI_SCHEME = "bitcoin"
logger = logging.getLogger(__name__)


class InvalidBitcoinURI(Exception):
    pass


def serialized_to_hex(serialized):
    return bytes(serialized).hex()


def hex_to_serialized(hex_string):
    return bytes.fromhex(hex_string)


class DecodingException(Exception):
    pass


class InconsistentDescriptors(Exception):
    pass


class WrongNetwork(Exception):
    pass


def _flatten_descriptors(descriptor: Descriptor) -> List[Descriptor]:
    if descriptor.pubkeys and len(descriptor.subdescriptors) == 0:
        return [descriptor]
    elif len(descriptor.pubkeys) == 0 and len(descriptor.subdescriptors) == 1:
        return [descriptor] + _flatten_descriptors(descriptor.subdescriptors[0])
    else:
        raise Exception(f"The descriptor cannot be flattened")



import json
import logging
import re
from typing import Dict, List, Optional, Tuple

import bdkpython as bdk
from hwilib.descriptor import Descriptor, PubkeyProvider, parse_descriptor

from bitcoin_qr_tools.utils import WrongNetwork, _flatten_descriptors

logger = logging.getLogger(__name__)


class SignerInfo:
    def __init__(
        self,
        fingerprint: str,
        key_origin: str,
        xpub: str,
        derivation_path: Optional[str] = None,
        name: Optional[str] = None,
        first_address: Optional[str] = None,
    ) -> None:
        """_summary_

        Args:
            fingerprint (str): Root fingerprint
            key_origin (str): _description_
            xpub (str): _description_
            derivation_path (Optional[str], optional): _description_. Defaults to None.
            name (Optional[str], optional): _description_. Defaults to None.
            first_address (Optional[str], optional): _description_. Defaults to None.
        """
        self.fingerprint = fingerprint
        self.key_origin = self.format_key_origin(key_origin)
        self.xpub = xpub
        self.derivation_path = derivation_path
        self.name = name
        self.first_address = first_address

    def format_key_origin(self, value):
        assert value.startswith("m"), "The value must start with m"
        return value.replace("'", "h")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def __str__(self) -> str:
        return f"{self.__dict__}"

    def to_json(self) -> str:
        return json.dumps(self.__dict__)

    def dict(self) -> Dict:
        return self.__dict__

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__

    @classmethod
    def _is_testnet(cls, network: bdk.Network) -> bool:
        return network in [
            bdk.Network.REGTEST,
            bdk.Network.TESTNET,
            bdk.Network.SIGNET,
        ]

    @classmethod
    def _inner_most_pubkey_provider(cls, descriptors: List[Descriptor]) -> Tuple[List[str], PubkeyProvider]:
        """
        Deturns the name of the descriptor (also when it was nested) and the single PubkeyProvider found

        If there are more than 1 PubkeyProvider present, it will raise DecodingException

        Args:
            descriptors (List[Descriptor]): _description_

        Raises:
            Exception: _description_

        Returns:
            Tuple[str, PubkeyProvider]: _description_
        """
        if len(descriptors) != 1:
            raise Exception("Only 1 descriptor expected")
        flattened_descriptors = _flatten_descriptors(descriptors[0])

        names = [d.name for d in flattened_descriptors]
        assert len(flattened_descriptors[-1].pubkeys) == 1
        pubkey_provider = flattened_descriptors[-1].pubkeys[0]
        return names, pubkey_provider

    @classmethod
    def hwi_pubkey_provider_to_signer_info(
        cls, hwi_pk_prov: PubkeyProvider, network: bdk.Network, name: str | None = None
    ) -> "SignerInfo":
        if hwi_pk_prov.extkey:
            is_testnet = hwi_pk_prov.extkey.is_testnet
            if is_testnet != cls._is_testnet(network):
                raise WrongNetwork(
                    f"""Expected Network {network}, but got {'Testnet' if is_testnet else 'Mainnet'}"""
                )

        return SignerInfo(
            fingerprint=hwi_pk_prov.origin.fingerprint.hex(),
            key_origin=hwi_pk_prov.origin.get_derivation_path(),
            xpub=hwi_pk_prov.pubkey,
            derivation_path=hwi_pk_prov.deriv_path,
            name=name,
        )

    @classmethod
    def _handle_incomplete_descritpor_bare_p2wsh(cls, descriptor_str: str) -> "Optional[SignerInfo]":
        if not (descriptor_str.startswith("wsh([") and "]" in descriptor_str):
            return None
        # handle_incorrect_output_descriptor_from_keystone
        # extract just the inner part
        names = ["wsh"]
        descriptor_str = descriptor_str.split("#")[0] if "#" in descriptor_str else descriptor_str
        inner_pubkey_str = descriptor_str.replace("wsh(", "").rstrip(")")
        try:
            signer_info = SignerInfo.from_str(inner_pubkey_str)
            signer_info.name = "-".join(["p2" + name for name in names])
        except:
            return None
        return signer_info

    @classmethod
    def decode_descriptor_as_signer_info(cls, descriptor_str: str, network: bdk.Network) -> "SignerInfo":
        try:
            descriptor = parse_descriptor(desc=descriptor_str)
            names, pubkey_provider = cls._inner_most_pubkey_provider([descriptor])
            return cls.hwi_pubkey_provider_to_signer_info(
                pubkey_provider, network=network, name="-".join(["p2" + name for name in names])
            )
        except ValueError as e:
            if "A matching pair of parentheses cannot be found" in e.args:
                signer_info = cls._handle_incomplete_descritpor_bare_p2wsh(descriptor_str=descriptor_str)
                if signer_info:
                    return signer_info
            raise e

    @classmethod
    def from_str(cls, s: str) -> "SignerInfo":
        """
        Splits 1 keystore,e.g. "[a42c6dd3/84'/1'/0']xpub/0/*"
        into fingerprint, key_origin, xpub, wallet_path

        It also replaces the "'" into "h"

        It overwrites fingerprint, key_origin, xpub  in default_keystore.
        """
        # one could use:
        #     pubkey_provider, _ = parse_pubkey(s)
        #     return SignerInfo(
        #         fingerprint=pubkey_provider.origin.fingerprint.hex(),
        #         key_origin=pubkey_provider.origin.get_derivation_path(),
        #         xpub=pubkey_provider.pubkey,
        #         derivation_path=pubkey_provider.deriv_path,
        #     )
        # however this cannot handle slip-132

        def key_origin_contains_valid_characters(s: str) -> bool:
            # Matches strings that consist of 'h', '/', digits, and optionally ends with a single quote
            return re.fullmatch("[mh/0-9']*", s) is not None

        def extract_groups(string: str, pattern):
            match = re.match(pattern, string)
            if match is None:
                raise Exception(f"'{string}' does not match the required pattern!")
            return match.groups()

        groups = extract_groups(s, r"\[(.*?)\/(.*?)\](.*?)(\/.*?)?$")

        key_origin = "m/" + groups[1].replace("'", "h")
        # guard against false positive detections
        assert key_origin_contains_valid_characters(key_origin)

        return SignerInfo(
            fingerprint=groups[0],
            key_origin=key_origin,
            xpub=groups[2],
            derivation_path=groups[3],
        )



import hashlib
import logging

import base58
import bdkpython as bdk

from bitcoin_qr_tools.utils import WrongNetwork

logger = logging.getLogger(__name__)


class ConverterXpub:
    # Mapping of SLIP-132 version bytes to BIP32 version bytes
    version_bytes_map = {
        bytes.fromhex("04b24746"): bytes.fromhex("0488b21e"),  # zpub to xpub
        bytes.fromhex("04b2430c"): bytes.fromhex("0488ade4"),  # zprv to xprv
        bytes.fromhex("049d7cb2"): bytes.fromhex("0488b21e"),  # ypub to xpub
        bytes.fromhex("049d7878"): bytes.fromhex("0488ade4"),  # yprv to xprv
        bytes.fromhex("0295b43f"): bytes.fromhex("0488b21e"),  # Ypub to xpub
        bytes.fromhex("0295b005"): bytes.fromhex("0488ade4"),  # Yprv to xprv
        bytes.fromhex("02aa7ed3"): bytes.fromhex("0488b21e"),  # Zpub to xpub
        bytes.fromhex("02aa7a99"): bytes.fromhex("0488ade4"),  # Zprv to xprv
        bytes.fromhex("045f1cf6"): bytes.fromhex("043587cf"),  # vpub to tpub
        bytes.fromhex("045f18bc"): bytes.fromhex("04358394"),  # vprv to tprv
        bytes.fromhex("044a5262"): bytes.fromhex("043587cf"),  # upub to tpub
        bytes.fromhex("044a4e28"): bytes.fromhex("04358394"),  # uprv to tprv
        bytes.fromhex("024289ef"): bytes.fromhex("043587cf"),  # Upub to tpub
        bytes.fromhex("024285b5"): bytes.fromhex("04358394"),  # Uprv to tprv
        bytes.fromhex("02575483"): bytes.fromhex("043587cf"),  # Vpub to tpub
        bytes.fromhex("02575048"): bytes.fromhex("04358394"),  # Vprv to tprv
    }

    @classmethod
    def xpub_matches_network(cls, xpub: str, network: bdk.Network) -> bool:
        if network == bdk.Network.BITCOIN:
            if not xpub.startswith("xpub"):
                return False
        else:
            if not xpub.startswith("tpub"):
                return False
        return True

    @classmethod
    def ensure_xpub_matches_network(cls, xpub: str, network: bdk.Network):
        "Raises an exception if not matching"
        if network == bdk.Network.BITCOIN:
            if not xpub.startswith("xpub"):
                raise WrongNetwork(f"{xpub} doesnt start with xpub, which is required for {network}")
        else:
            if not xpub.startswith("tpub"):
                raise WrongNetwork(f"{xpub} doesnt start with tpub, which is required for {network}")

    @classmethod
    def is_xpub(cls, s: str) -> bool:
        if not s.isalnum():
            return False
        first_four_letters = s[:4]
        return first_four_letters.endswith("pub")

    ################ here is the slip132 part
    ### see https://github.com/satoshilabs/slips/blob/master/slip-0132.md

    @classmethod
    def get_slip132_version_bytes(cls, slip132_key: str) -> bytes:
        """Get the version bytes from a SLIP-132 key."""
        raw_extended_key = base58.b58decode(slip132_key)
        return raw_extended_key[:4]

    @classmethod
    def base58check_decode(cls, s: str) -> bytes:
        """Decode a Base58Check encoded string to bytes"""
        # Decode the string
        data = base58.b58decode(s)

        # Split the data into the payload and the checksum
        check_sum = data[-4:]
        payload = data[:-4]

        # Calculate the checksum of the payload
        calculated_check_sum = hashlib.sha256(hashlib.sha256(payload).digest()).digest()[:4]

        # Compare the calculated checksum with the checksum from the data
        if check_sum != calculated_check_sum:
            raise ValueError("Invalid checksum")

        return payload

    @classmethod
    def normalized_to_bip32(cls, xpub: str) -> str:
        if ConverterXpub.is_slip132(xpub):
            converted = cls.convert_slip132_to_bip32(xpub)
            logger.debug(f"Converting SLIP132 to Bip32: {xpub} -> {converted}")
            return converted
        return xpub

    @classmethod
    def convert_slip132_to_bip32(cls, slip132_key: str) -> str:
        """Convert a SLIP-132 extended key to a BIP32 extended key."""
        raw_extended_key = cls.base58check_decode(slip132_key)
        slip132_version_bytes = raw_extended_key[:4]

        # Lookup the corresponding BIP32 version bytes
        bip32_version_bytes = cls.version_bytes_map.get(slip132_version_bytes)
        if bip32_version_bytes is None:
            raise ValueError("Unsupported SLIP-132 version bytes")

        # Replace the version bytes of the raw key
        replaced_version_key = bip32_version_bytes + raw_extended_key[4:]

        # Calculate the checksum of the replaced version key
        check_sum = hashlib.sha256(hashlib.sha256(replaced_version_key).digest()).digest()[:4]

        # Encode the replaced version key + checksum into Base58
        bip32_key = base58.b58encode(replaced_version_key + check_sum).decode()

        return bip32_key

    @classmethod
    def is_slip132(cls, key: str) -> bool:
        try:
            return cls.get_slip132_version_bytes(key) in cls.version_bytes_map
        except:
            return False



import mss
import numpy as np
import pygame


class ScreenCamera:
    def __init__(self):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def _numpy_to_surface(self, numpy_image: np.ndarray) -> pygame.Surface:
        # Convert numpy image (RGB) to pygame surface
        return pygame.surfarray.make_surface(numpy_image.transpose((1, 0, 2)))

    def get_image(self):
        with mss.mss() as sct:
            monitor = sct.monitors[1]  # capture the first monitor
            sct_img = sct.grab(monitor)

        image = np.array(sct_img)

        # Convert BGRA to RGB by reordering the channels
        image = image[:, :, [2, 1, 0]]  # Rearrange BGR to RGB

        # Convert numpy image to surface for drawing
        surface = self._numpy_to_surface(image)
        surface = pygame.transform.flip(surface, True, False)
        return surface

    def __del__(self):
        self.stop()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import sys
from pathlib import Path
from typing import List, Optional, Tuple

from PIL import Image
from PyQt6.QtCore import QByteArray, QEvent, QObject, QRectF, QSize, Qt, QTimer
from PyQt6.QtGui import (
    QCloseEvent,
    QEnterEvent,
    QIcon,
    QImage,
    QKeyEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QPixmap,
)
from PyQt6.QtSvg import QSvgRenderer
from PyQt6.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from bitcoin_qr_tools.data import Data
from bitcoin_qr_tools.qr_generator import QRGenerator
from bitcoin_qr_tools.unified_encoder import QrExportTypes, UnifiedEncoder

logger = logging.getLogger(__name__)


def resource_path(*parts: str):
    pkg_dir = os.path.split(os.path.realpath(__file__))[0]
    return os.path.join(pkg_dir, *parts)


def icon_path(icon_basename: str):
    return resource_path("gui", "icons", icon_basename)


def pil_image_to_qimage(im: Image.Image):
    im = im.convert("RGBA")
    data = im.tobytes("raw", "RGBA")
    qim = QImage(data, im.size[0], im.size[1], QImage.Format.Format_RGBA8888)

    return qim.copy()  # Making a copy to let data persist after function returns


class ImageWidget(QWidget):
    def __init__(
        self, pil_image: Image.Image | None = None, parent=None, size_hint: Tuple[int, int] | None = None
    ):
        super().__init__(parent)
        self.pil_image = pil_image
        self.size_hint = size_hint
        self.qt_image = pil_image_to_qimage(pil_image) if pil_image else QImage()
        self.scaled_image = self.qt_image

    def paintEvent(self, event: QPaintEvent | None) -> None:
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, False)

        if not self.qt_image.isNull():
            widget_width, widget_height = self.width(), self.height()

            # choose minimum of image and self sizes
            width = min(self.qt_image.size().width(), widget_width)
            height = min(self.qt_image.size().height(), widget_height)

            # Scale the image to fit within the widget while maintaining aspect ratio
            self.scaled_image = self.qt_image.scaled(
                width, height, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation
            )

            # Calculate position to center the image
            x = (widget_width - self.scaled_image.width()) // 2
            y = (widget_height - self.scaled_image.height()) // 2

            # Draw the image centered
            painter.drawImage(x, y, self.scaled_image)

    def set_image(self, pil_image: Image.Image):
        self.pil_image = pil_image
        self.qt_image = pil_image_to_qimage(pil_image)
        self.update()  # Trigger a repaint

    def load_from_file(self, filepath: str):
        self.set_image(Image.open(filepath))

    def sizeHint(self) -> QSize:
        if not self.qt_image.isNull():
            if not self.size_hint:
                return self.qt_image.size()
            else:
                s = QSize()
                s.setWidth(self.size_hint[0])
                s.setHeight(self.size_hint[1])
                return s
        return super().sizeHint()


class EnlargableImageWidget(ImageWidget):
    def __init__(
        self, pil_image: Image.Image | None = None, parent=None, size_hint: Tuple[int, int] | None = None
    ):
        super().__init__(pil_image, parent, size_hint=size_hint)
        self.enlarged_image: Optional[EnlargedImage] = None
        self.setCursor(Qt.CursorShape.PointingHandCursor)

    def enlarge_image(self):
        if not self.enlarged_image:
            self.enlarged_image = EnlargedImage(self.pil_image)

        if self.enlarged_image.isVisible():
            self.enlarged_image.close()
        else:
            self.enlarged_image.show()

    def mousePressEvent(self, event: QMouseEvent | None):
        self.enlarge_image()
        super().mousePressEvent(event)

    def closeEvent(self, event: QCloseEvent | None) -> None:
        if self.enlarged_image:
            self.enlarged_image.close()
        super().closeEvent(event)


class EnlargedImage(ImageWidget):
    def __init__(self, pil_image: Image.Image, parent=None, screen_fraction=0.7):
        super().__init__(pil_image, parent)
        self.setWindowTitle("Enlarged Image")
        self.installEventFilter(self)  # Install the event filter for this widget

        # Get screen resolution
        screen = QApplication.screens()[0].size()

        # Calculate the new size maintaining the aspect ratio
        image_aspect_ratio = self.qt_image.width() / self.qt_image.height()
        new_width = min(screen.width() * screen_fraction, self.qt_image.width())
        new_height = new_width / image_aspect_ratio

        # Ensure the height does not exceed 50% of the screen height
        if new_height > screen.height() * screen_fraction:
            new_height = screen.height() * screen_fraction
            new_width = new_height * image_aspect_ratio

        # Calculate position to center the window
        x = round((screen.width() - new_width) / 2)
        y = round((screen.height() - new_height) / 2)

        self.setGeometry(x, y, round(new_width), round(new_height))

    def eventFilter(self, source: QObject | None, event: QEvent | None) -> bool:
        # Check for the FocusOut event
        if event and event.type() in [QEvent.Type.FocusOut, QEvent.Type.WindowDeactivate]:
            # Close the widget if it loses focus
            if source is self:
                self.close()
        return super().eventFilter(source, event)

    def mousePressEvent(self, event: QMouseEvent | None):
        self.close()

    def keyPressEvent(self, event: QKeyEvent | None):
        if event and event.key() == Qt.Key.Key_Escape:
            self.close()


class EnlargableImageWidgetWithButton(QWidget):
    def __init__(
        self,
        pil_image: Image.Image | None = None,
        parent: QWidget | None = None,
        size_hint: Tuple[int, int] | None = None,
    ):
        super().__init__(parent)

        layout = QVBoxLayout(self)
        self.enlargable_widget = EnlargableImageWidget(pil_image=pil_image, parent=self, size_hint=size_hint)

        layout.addWidget(self.enlargable_widget)

        self.button_enlarge_qr = QPushButton()
        self.button_enlarge_qr.setIcon(QIcon(icon_path("zoom.png")))
        # self.button_enlarge_qr.setIconSize(QSize(30, 30))  # 24x24 pixels
        self.button_enlarge_qr.clicked.connect(self.enlargable_widget.enlarge_image)
        self.button_enlarge_qr.setText(self.tr("Enlarge"))
        self.button_enlarge_qr.setMaximumWidth(200)
        layout.addWidget(self.button_enlarge_qr, alignment=Qt.AlignmentFlag.AlignHCenter)

    def load_from_file(self, filepath: str):
        self.enlargable_widget.load_from_file(filepath=filepath)


class QRCodeWidget(EnlargableImageWidget):
    def __init__(self, parent=None, clickable=True):
        super().__init__(parent=parent)
        # QR code specific initializations, if any

    def set_data(self, data: str):
        # Implement QR code generation and setting image
        self.set_image(QRGenerator.create_qr_PILimage(data))


#######################################
# svg widgets


class QRCodeWidgetSVG(QWidget):
    def __init__(self, always_animate=False, clickable=True, parent=None):
        super().__init__(parent)
        self.svg_renderers: List[QSvgRenderer] = []
        self.current_index = 0
        self.enlarged_image = None
        self.clickable = clickable
        self.always_animate = always_animate
        self.is_hovered = False
        self.default_size = 200
        self.setBaseSize(self.default_size, self.default_size)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.setMinimumSize(20, 20)

        if clickable:
            self.setCursor(Qt.CursorShape.PointingHandCursor)

        self.timer = QTimer()
        self.timer.timeout.connect(self.next_svg)

    def set_renderers(self, svg_renderers):
        self.svg_renderers = svg_renderers
        self.current_index = 0
        if self.enlarged_image:
            if self.svg_renderers:
                self.enlarged_image.update_image(self.svg_renderers[self.current_index])
            else:
                self.enlarged_image.update_image(None)
        self.manage_animation()
        # if i update, when it is not visible, then I force the window to show
        if self.isVisible():
            self.update()

    def set_data_list(self, data_list: List[str]):
        self.set_renderers(
            [QSvgRenderer(QByteArray(QRGenerator.create_qr_svg(data).encode("utf-8"))) for data in data_list]
        )

    def set_always_animate(self, always_animate: bool):
        self.always_animate = always_animate
        self.manage_animation()

    def set_images(self, image_list: List[str]):
        self.set_renderers([QSvgRenderer(QByteArray(image.encode("utf-8"))) for image in image_list])

    def manage_animation(self):
        should_animate = len(self.svg_renderers) > 1 and (
            self.always_animate
            or self.is_hovered
            or (self.enlarged_image and self.enlarged_image.isVisible())
        )
        if should_animate:
            self.timer.start(1000)  # Change SVG every 1 second
        else:
            self.timer.stop()

    def next_svg(self):
        if not self.svg_renderers:
            return

        self.current_index = (self.current_index + 1) % len(self.svg_renderers)
        self.update()
        if self.enlarged_image and self.enlarged_image.isVisible():
            self.enlarged_image.update_image(self.svg_renderers[self.current_index])
        else:
            self.manage_animation()

    def paintEvent(self, event: QPaintEvent | None) -> None:
        if not self.svg_renderers:
            return

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        widget_width, widget_height = self.width(), self.height()
        side = min(widget_width, widget_height)
        x = (widget_width - side) // 2
        y = (widget_height - side) // 2

        self.svg_renderers[self.current_index].render(painter, QRectF(x, y, side, side))

    def enterEvent(self, event: QEnterEvent | None) -> None:
        self.is_hovered = True
        self.manage_animation()
        super().enterEvent(event)

    def leaveEvent(self, event: QEvent | None) -> None:
        self.is_hovered = False
        self.manage_animation()
        super().leaveEvent(event)

    def enlarge_image(self):
        if not self.svg_renderers:
            return

        if not self.enlarged_image:
            self.is_hovered = False
            self.enlarged_image = EnlargedSVG(self.svg_renderers[self.current_index])

        self.enlarged_image.exec()
        self.enlarged_image.update_image(self.svg_renderers[self.current_index])
        self.manage_animation()

    def mousePressEvent(self, event: QMouseEvent | None):
        if self.clickable:
            self.enlarge_image()
        super().mousePressEvent(event)

    def save_file(self, filename: Path, antialias=False):
        """Save all QR codes to files. If format is 'GIF', combines them into
        an animated GIF.

        :param filename:
        :param antialias: Boolean to indicate if anti-aliasing should be
            used.
        """
        if len(self.svg_renderers) > 1:
            images: List[Image.Image] = []
            for renderer in self.svg_renderers:
                if not renderer.isValid():
                    continue
                images.append(self.renderer_to_pil(renderer, antialias))
            images[0].save(
                str(filename),
                format="gif",
                save_all=True,
                append_images=images[1:],
                loop=0,
                duration=1000,
            )
        elif len(self.svg_renderers) == 1:
            renderer = self.svg_renderers[0]
            if renderer.isValid():
                image = self.renderer_to_pil(renderer, antialias)
                image.save(str(filename))

    def renderer_to_pil(self, renderer: QSvgRenderer, antialias: bool) -> Image.Image:
        """Convert a QR code renderer to a PIL Image.

        :param renderer: The QR code renderer.
        :param antialias: Boolean to indicate if anti-aliasing should be
            used.
        :return: PIL Image object.
        """
        # setting the length that it can handle the largets qr codes:
        # Version 40: 177x177 modules
        length = 177 * 2

        pixmap = QPixmap(length, length)
        pixmap.fill(Qt.GlobalColor.white)
        painter = QPainter(pixmap)

        if antialias:
            painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        renderer.render(painter, QRectF(0, 0, length, length))
        painter.end()

        # Convert QPixmap to QImage
        qimage = pixmap.toImage()

        return self.qimage_to_pil(qimage=qimage)

    @staticmethod
    def qimage_to_pil(qimage: QImage) -> Image.Image:
        """Convert QImage to PIL Image."""
        qimage = qimage.convertToFormat(QImage.Format.Format_RGB32)

        width = qimage.width()
        height = qimage.height()

        # Get the data from the QImage
        # 4 bytes per pixel
        buffer = qimage.bits().asstring(width * height * 4)  # type: ignore

        # Create a PIL Image from the buffer
        pil_image = Image.frombytes("RGBA", (width, height), buffer, "raw", "BGRA")

        return pil_image

    def as_pil_images(self):
        """Convert all the QR codes to PIL Images.

        :return: List of PIL Image objects.
        """
        return [
            self.renderer_to_pil(renderer, antialias=True)
            for renderer in self.svg_renderers
            if renderer.isValid()
        ]

    def closeEvent(self, event: QCloseEvent | None) -> None:
        self.timer.stop()
        if self.enlarged_image:
            self.enlarged_image.close()
        super().closeEvent(event)


class EnlargedSVG(QDialog):
    def __init__(self, svg_renderer: QSvgRenderer, parent=None, screen_fraction=0.5):
        super().__init__(parent)
        self.svg_renderer = svg_renderer
        self.setWindowTitle("QR Code")
        # self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.WindowStaysOnTopHint)
        screen = QApplication.screens()[0].size()
        width = height = round(min(screen.width(), screen.height()) * screen_fraction)
        self.setGeometry(
            round((screen.width() - width) / 2),
            round((screen.height() - height) / 2),
            width,
            height,
        )

    def update_image(self, new_renderer: QSvgRenderer):
        self.svg_renderer = new_renderer
        self.update()

    def paintEvent(self, event: QPaintEvent | None) -> None:
        if not self.svg_renderer:
            return

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        widget_width, widget_height = self.width(), self.height()
        side = min(widget_width, widget_height)
        x = (widget_width - side) // 2
        y = (widget_height - side) // 2

        self.svg_renderer.render(painter, QRectF(x, y, side, side))

    def mousePressEvent(self, event: QMouseEvent | None):
        super().mousePressEvent(event)
        self.close()

    def keyPressEvent(self, event: QKeyEvent | None):
        if event and event.key() == Qt.Key.Key_Escape:
            self.close()


if __name__ == "__main__":
    import bdkpython as bdk

    testdata = [
        Data.from_str(
            "cHNidP8BAHEBAAAAAXgQzjk+DTWQTPUtRMbYiheC0jfbipvw+jQ5lidmyABjAAAAAAD9////AgDh9QUAAAAAFgAUbBuOQOlcnz8vpruh2Kb3CFr4vlhkEQ2PAAAAABYAFN1n2hvBWYzshD42xwQzy9XYoji3BAEAAAABAKoCAAAAAAEBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAD/////BQKYAAEB/////wIA+QKVAAAAABYAFLlHwN6VXNLM381bMxmNJlaDTQzVAAAAAAAAAAAmaiSqIant4vYcP3HR3v0/qZnfo2lTdVxpBol5mWK0i+vYNpdOjPkBIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEBHwD5ApUAAAAAFgAUuUfA3pVc0szfzVszGY0mVoNNDNUiBgISCnRxeOxzC0MgK01AmiIRLrgS1AyIqKeBkdwL+nt/6RikLG3TVAAAgAEAAIAAAACAAAAAAAAAAAAAACICAlQcwExiTUk9f7olLkwPlQpiregRHc9jXXFJBlMoucgNGKQsbdNUAACAAQAAgAAAAIAAAAAAAQAAAAA=",
            network=bdk.Network.REGTEST,
        ),
        Data.from_str(
            "wsh(sortedmulti(2,[829074ff/48'/1'/0'/2']tpubDDx9arPwEvHGnnkKN1YJXFE4W6JZXyVX9HGjZW75nWe1FCsTYu2k3i7VtCwhGR9zj6UUYnseZUnwL7T6Znru3NmXkcjEQxMqRx7Rxz8rPp4/<0;1>/*,[45f35351/48'/1'/0'/2']tpubDEY3tNWvDs8J6xAmwoirxgff61gPN1V6U5numeb6xjvZRB883NPPpRYHt2A6fUE3YyzDLezFfuosBdXsdXJhJUcpqYWF9EEBmWqG3rG8sdy/<0;1>/*,[d5b43540/48'/1'/0'/2']tpubDFnCcKU3iUF4sPeQC68r2ewDaBB7TvLmQBTs12hnNS8nu6CPjZPmzapp7Woz6bkFuLfSjSpg6gacheKBaWBhDnEbEpKtCnVFdQnfhYGkPQF/<0;1>/*))#2jxldwxn",
            network=bdk.Network.REGTEST,
        ),
        Data.from_str(
            "wpkh([a42c6dd3/84h/1h/0h]tpubDDnGNapGEY6AZAdQbfRJgMg9fvz8pUBrLwvyvUqEgcUfgzM6zc2eVK4vY9x9L5FJWdX8WumXuLEDV5zDZnTfbn87vLe9XceCFwTu9so9Kks/0/*)",
            network=bdk.Network.REGTEST,
        ),
        Data.from_str(
            "0100000000010177ff6b4de45caf689a95367958ff6b912c2385d4d7563a09ba41cb0a2c30f5220000000000fdffffff02a0cee90100000000160014f22e4b1c92222a38b286fdd39ee2e35d4b581c47d62019930000000016001412c9e0c94dd6c71cfec7bcff16feea0a0fb8bc5d0247304402205c88d4d7e3059f16c6f74debb9754efeba89f92a237b7d09d9732e59b8a7d6de02202ce0ef338af77ebd8c14ae88f7e83116e0ba27a89aee7829ef70d1fc8d99af06012102802e1fda05b62b1f071d35bcd129fc0f9cf3517c6af7b3bb0ce76d76c7de068d00000000",
            network=bdk.Network.REGTEST,
        ),
        Data.from_str(
            "[a42c6dd3/84'/1'/0']tpubDDnGNapGEY6AZAdQbfRJgMg9fvz8pUBrLwvyvUqEgcUfgzM6zc2eVK4vY9x9L5FJWdX8WumXuLEDV5zDZnTfbn87vLe9XceCFwTu9so9Kks/0/*",
            network=bdk.Network.REGTEST,
        ),
        Data.from_str(
            '{"chain": "XRT", "xfp": "0F056943", "p2sh": {"xpub": "tpubD8NXmKsmWp3a3DXhbihAYbYLGaRNVdTnr6JoSxxfXYQcmwVtW2hv8QoDwng6JtEonmJoL3cNEwfd2cLXMpGezwZ2vL2dQ7259bueNKj9C8n", "deriv": "m/45h", "first": null, "name": "p2sh"}, "p2sh_p2wsh": {"xpub": "tpubDF2rnouQaaYrUEy2JM1YD3RFzew4onawGM4X2Re67gguTf5CbHonBRiFGe3Xjz7DK88dxBFGf2i7K1hef3PM4cFKyUjcbJXddaY9F5tJBoP", "deriv": "m/48h/1h/0h/1h", "first": null, "name": "p2sh_p2wsh"}, "p2wsh": {"xpub": "tpubDF2rnouQaaYrXF4noGTv6rQYmx87cQ4GrUdhpvXkhtChwQPbdGTi8GA88NUaSrwZBwNsTkC9bFkkC8vDyGBVVAQTZ2AS6gs68RQXtXcCvkP", "deriv": "m/48h/1h/0h/2h", "first": null, "name": "p2wsh"}}',
            network=bdk.Network.REGTEST,
        ),
    ]

    class ComboBoxDemo(QWidget):
        def __init__(self):
            super().__init__()

            # Setup the layout
            layout = QVBoxLayout()

            # enlarable_widget2 = EnlargableImageWidgetWithButton()
            # enlarable_widget2.load_from_file("docs/bad-light.png")
            # layout.addWidget(enlarable_widget2)

            self.svg_widget = QRCodeWidgetSVG()
            self.svg_widget.set_always_animate(True)
            layout.addWidget(self.svg_widget)

            # Initialize combo box 1
            self.combo_data = QComboBox()
            for data in testdata:
                self.combo_data.addItem(data.data_type.name, data)
            self.combo_data.currentIndexChanged.connect(self.on_combo_changed)
            layout.addWidget(self.combo_data)

            # Initialize combo box 2
            self.combo_qrtype = QComboBox()
            for qr_type in QrExportTypes.as_list():
                self.combo_qrtype.addItem(qr_type.display_name, qr_type)
            self.combo_qrtype.currentIndexChanged.connect(self.on_combo_changed)
            layout.addWidget(self.combo_qrtype)

            # Set the layout to the QWidget
            self.setLayout(layout)

            self.on_combo_changed()

        def on_combo_changed(self):
            # Get current data from both combo boxes
            data = self.combo_data.currentData()
            qrtype = self.combo_qrtype.currentData()

            fragments = UnifiedEncoder.generate_fragments_for_qr(data=data, qr_export_type=qrtype)
            self.svg_widget.set_data_list(fragments)

    app = QApplication(sys.argv)
    window = ComboBoxDemo()
    window.show()
    sys.exit(app.exec())
